﻿local debug = false;

function LKA_debug(message)
  if(debug) then
    LKA_message("|cFF00FFFFDebug|r|cFFFFFF00 - "..message);
  end
end

function LKA_format_number(number)
  local formatted = (number > 100000 and format("%.1f", number/1000000).."M") or (number > 1000 and format("%i", number/1000)..","..(number%1000)) or number;
  return formatted;
end

function LKA_ratio_to_hex_color(r, g, b)
  local red = r * 255;
  local green = g * 255;
  local blue = b * 255;
  return LKA_dec_to_hex_color(red, green, blue);
end

function LKA_dec_to_hex_color(r, g, b)
  local red = LKA_pad(string.format("%X", r));
  local green = LKA_pad(string.format("%X", g));
  local blue = LKA_pad(string.format("%X", b));
  return red..green..blue
end

function LKA_pad(input)
    if(string.len(input) == 1) then
        return "0"..input
    elseif(string.len(input) == 0) then
        return "00"
    else
        return input
    end
end

function LKA_message(message)
  print("|cFFFF0000LKAnnounce|r|cFFFFFF00: "..message);
end

function LKA_printraid(message)
  if(GetNumRaidMembers() > 0) then
    SendChatMessage(message, "RAID", nil, nil);
  else
    LKA_message(message);
  end
end


local ADDON_NAME = "LKAnnounce";
local LKAnnounce = LibStub("AceAddon-3.0"):NewAddon(ADDON_NAME);
local talentLib = LibStub and LibStub:GetLibrary("LibGroupTalents-1.0", true)
local LDB = LibStub:GetLibrary("LibDataBroker-1.1");

local frame;
local pending_defile = false;
local cooldowns = {};
local phase = 1;
local raid_positions = {};
local infest_count = 1;
local trap_targets = {};
local picked = {};
local num_picked = 0;

local horror_names = {};
local current_horror = 1;
local plague_announced = {};
local targeting_horrors = {};

local NONE = " none";
local NONE_DISPLAY = "|cFFFF0000None";

local NORAID = "|cFFFF0000Not in a Raid Group|r";
local loaded_header = NORAID;

local ZONE = "Icecrown Citadel";
local SUBZONE = "The Frozen Throne";
local LKPULL = "So the Light's vaunted justice has finally arrived? Shall I lay down Frostmourne and throw myself at your mercy, Fordring?";
local encounter_enabled = false;

local ZOOMED_ICON = ":16:16:0:0:64:64:4:60:4:60";
local ZOOMED_ICON_SMALL = ":12:12:0:0:64:64:4:60:4:60";
local RAID_TARGET = ":14:14:0:0:64:64";

local ASSIGNMENT_HELP = [[|cFF3030FFVal'kyr Assignments|r

Some possible stuns/slows/infests will not be available until talents are loaded.

Backup and aoe stuns and slows, and infest cooldowns, are assigned in a priority - the first assignment is chosen unless he/she is disabled, in which case the second assignment is chosen, etc.

Single target stuns can be assigned in rotation - if the main assignment is longer than 40s cooldown, the second assignment will always be assigned to backup. Leaving the second slot blank will cause a backup to be picked from the regular backups list]]

local channel_colors = {
  RAID = "FFFFFF",
  RAID_WARNING = "FFFFFF",
  YELL = "FFFFFF",
  PARTY = "FFFFFF",
  SAY = "FFFFFF",
  GUILD = "FFFFFF"
}
local channels = {
  RAID = "Raid",
  RAID_WARNING = "cRW",
  YELL = "Yell",
  PARTY = "Party",
  SAY = "Say",
  GUILD = "Guild"
}

local class_colors = {
  ["DEATHKNIGHT"] = "|cFFC41F3B",
  ["DRUID"] = "|cFFFF7D0A",
  ["HUNTER"] = "|cFFABD473",
  ["MAGE"] = "|cFF69CCF0",
  ["PALADIN"] = "|cFFF58CBA",
  ["PRIEST"] = "|cFFFFFFFF",
  ["ROGUE"] = "|cFFFFF569",
  ["SHAMAN"] = "|cFF2459FF",
  ["WARLOCK"] = "|cFF9482C9",
  ["WARRIOR"] = "|cFFC79C6E",
  gray = "|cFFD0D0D0"
}

local getNames = function(self, assignment, index)
  local ret = {};
  for i,v in pairs(self.content) do
    ret[i] = v.name;
  end
  
  --Mark all elements in use by others in the assignment as used
  for i,v in pairs(LKAnnounceSaved.valks[assignment]) do
    if(self.content[v] and i ~= index) then
      ret[v] = self.content[v].inUseName;
    end
  end
  
  --If the given index's assignment is not in the content list (that assignment is offline or incorrectly specced), insert it but mark it as disabled
  if (LKAnnounceSaved.valks[assignment][index] and not self.content[LKAnnounceSaved.valks[assignment][index]]) then
    ret[LKAnnounceSaved.valks[assignment][index]] = class_colors.gray..LKAnnounceSaved.valks[assignment][index];
  end
  
  return ret;
end

local stuns = {
  names = getNames,
  content = {
    [NONE] = {
      name = NONE_DISPLAY,
      inUseName = NONE_DISPLAY
    }
  }
};
local rotation_stuns = {
  names = getNames,
  content = {
    [NONE] = {
      name = NONE_DISPLAY,
      inUseName = NONE_DISPLAY
    }
  }
};
local slows = {
  names = getNames,
  content = {
    [NONE] = {
      name = NONE_DISPLAY,
      inUseName = NONE_DISPLAY
    }
  }
};
local aoe_stuns = {
  names = getNames,
  content = {
    [NONE] = {
      name = NONE_DISPLAY,
      inUseName = NONE_DISPLAY
    }
  }
};
local aoe_slows = {
  names = getNames,
  content = {
    [NONE] = {
      name = NONE_DISPLAY,
      inUseName = NONE_DISPLAY
    }
  }
};
local infests = {
  names = getNames,
  content = {
    [NONE] = {
      name = NONE_DISPLAY,
      inUseName = NONE_DISPLAY
    }
  }
};
local num_infests = 0;

local shapes = {
  names = function(self)
    local ret = {};
    for i,v in pairs(self.content) do
      ret[i] = v.name;
    end
    return ret;
  end,
  content = {
    [0] = {
      name = "|TInterface\\Icons\\INV_Misc_QuestionMark"..ZOOMED_ICON.."|t Unknown",
      texture = "|TInterface\\Icons\\INV_Misc_QuestionMark"..ZOOMED_ICON.."|t",
      chatName = "?"
    },
    {
      name = "|TInterface\\TARGETINGFRAME\\UI-RaidTargetingIcon_1"..RAID_TARGET.."|t Star",
      texture = "|TInterface\\TARGETINGFRAME\\UI-RaidTargetingIcon_1"..RAID_TARGET.."|t",
      chatName = "{star}"
    },
    {
      name = "|TInterface\\TARGETINGFRAME\\UI-RaidTargetingIcon_2"..RAID_TARGET.."|t Circle",
      texture = "|TInterface\\TARGETINGFRAME\\UI-RaidTargetingIcon_2"..RAID_TARGET.."|t",
      chatName = "{circle}"
    },
    {
      name = "|TInterface\\TARGETINGFRAME\\UI-RaidTargetingIcon_3"..RAID_TARGET.."|t Diamond",
      texture = "|TInterface\\TARGETINGFRAME\\UI-RaidTargetingIcon_3"..RAID_TARGET.."|t",
      chatName = "{diamond}"
    },
    {
      name = "|TInterface\\TARGETINGFRAME\\UI-RaidTargetingIcon_4"..RAID_TARGET.."|t Triangle",
      texture = "|TInterface\\TARGETINGFRAME\\UI-RaidTargetingIcon_4"..RAID_TARGET.."|t",
      chatName = "{triangle}"
    },
    {
      name = "|TInterface\\TARGETINGFRAME\\UI-RaidTargetingIcon_5"..RAID_TARGET.."|t Moon",
      texture = "|TInterface\\TARGETINGFRAME\\UI-RaidTargetingIcon_5"..RAID_TARGET.."|t",
      chatName = "{moon}"
    },
    {
      name = "|TInterface\\TARGETINGFRAME\\UI-RaidTargetingIcon_6"..RAID_TARGET.."|t Square",
      texture = "|TInterface\\TARGETINGFRAME\\UI-RaidTargetingIcon_6"..RAID_TARGET.."|t",
      chatName = "{square}"
    },
    {
      name = "|TInterface\\TARGETINGFRAME\\UI-RaidTargetingIcon_7"..RAID_TARGET.."|t Cross",
      texture = "|TInterface\\TARGETINGFRAME\\UI-RaidTargetingIcon_7"..RAID_TARGET.."|t",
      chatName = "{cross}"
    },
    {
      name = "|TInterface\\TARGETINGFRAME\\UI-RaidTargetingIcon_8"..RAID_TARGET.."|t Skull",
      texture = "|TInterface\\TARGETINGFRAME\\UI-RaidTargetingIcon_8"..RAID_TARGET.."|t",
      chatName = "{skull}"
    }
  }
};

local shapes_sets = {
  names = function(self)
    local ret = {};
    for i,v in pairs(self.content) do
      ret[i] = v.name;
    end
    return ret;
  end,
  content = {
    BigWigs = {
      modName = "Big Wigs",
      name = "Big Wigs:  "..shapes.content[0].texture.." "..shapes.content[0].texture.." "..shapes.content[0].texture,
      marks = {0, 0, 0}
    },
    DBM = {
      modName = "Deadly Boss Mods",
      name = "DBM:  "..shapes.content[4].texture.." "..shapes.content[3].texture.." "..shapes.content[2].texture,
      marks = {4, 3, 2}
    },
    DXE = {
      shortName = "Deus Vox Encounters",
      name = "DXE:  "..shapes.content[5].texture.." "..shapes.content[6].texture.." "..shapes.content[7].texture,
      marks = {5, 6, 7}
    },
    LKAnnounce = {
      name = "LKAnnounce:  "..shapes.content[1].texture.." "..shapes.content[2].texture.." "..shapes.content[3].texture,
      shortName = "LKAnnounce",
      marks = {1, 2, 3}
    },
    RaidWatch = {
      shortName = "Raid Watch 2",
      name = "Raid Watch 2:  "..shapes.content[4].texture.." "..shapes.content[5].texture.." "..shapes.content[3].texture,
      marks = {4, 5, 6}
    }
  }
}

local possible_roles = {
  ["DEATHKNIGHT"] = {
    ["Unknown"] = "|TInterface\\Icons\\INV_Misc_QuestionMark"..ZOOMED_ICON.."|t",
    ["Frost"] = "|TInterface\\Icons\\Spell_DeathKnight_FrostPresence"..ZOOMED_ICON.."|t",
    ["Blood"] = "|TInterface\\Icons\\Spell_DeathKnight_BloodPresence"..ZOOMED_ICON.."|t",
    ["Unholy"] = "|TInterface\\Icons\\Spell_DeathKnight_UnholyPresence"..ZOOMED_ICON.."|t"
  },
  ["DRUID"] = {
    ["Unknown"] = "|TInterface\\Icons\\INV_Misc_QuestionMark"..ZOOMED_ICON.."|t",
    ["Balance"] = "|TInterface\\Icons\\Spell_Nature_StarFall"..ZOOMED_ICON.."|t",
    --["Feral Combat"] = "|TInterface\\Icons\\Ability_Racial_BearForm"..ZOOMED_ICON.."|t",
    ["Feral Combat"] = "|TInterface\\Icons\\Ability_Druid_CatForm"..ZOOMED_ICON.."|t",
    ["Restoration"] = "|TInterface\\Icons\\Spell_Nature_HealingTouch"..ZOOMED_ICON.."|t"
  },
  ["HUNTER"] = {
    ["Unknown"] = "|TInterface\\Icons\\INV_Misc_QuestionMark"..ZOOMED_ICON.."|t",
    ["Beast Mastery"] = "|TInterface\\Icons\\Ability_Hunter_BeastTaming"..ZOOMED_ICON.."|t",
    ["Survival"] = "|TInterface\\Icons\\Ability_Hunter_SwiftStrike"..ZOOMED_ICON.."|t",
    ["Marksmanship"] = "|TInterface\\Icons\\Ability_Marksmanship"..ZOOMED_ICON.."|t"
  },
  ["MAGE"] = {
    ["Unknown"] = "|TInterface\\Icons\\INV_Misc_QuestionMark"..ZOOMED_ICON.."|t",
    ["Frost"] = "|TInterface\\Icons\\spell_frost_frostbolt02"..ZOOMED_ICON.."|t",
    ["Fire"] = "|TInterface\\Icons\\spell_fire_flamebolt"..ZOOMED_ICON.."|t",
    ["Arcane"] = "|TInterface\\Icons\\spell_holy_magicalsentry"..ZOOMED_ICON.."|t"
  },
  ["PALADIN"] = {
    ["Unknown"] = "|TInterface\\Icons\\INV_Misc_QuestionMark"..ZOOMED_ICON.."|t",
    ["Holy"] = "|TInterface\\Icons\\spell_holy_holybolt"..ZOOMED_ICON.."|t",
    ["Protection"] = "|TInterface\\Icons\\spell_holy_devotionaura"..ZOOMED_ICON.."|t",
    ["Retribution"] = "|TInterface\\Icons\\spell_holy_auraoflight"..ZOOMED_ICON.."|t"
  },
  ["PRIEST"] = {
    ["Unknown"] = "|TInterface\\Icons\\INV_Misc_QuestionMark"..ZOOMED_ICON.."|t",
    ["Discipline"] = "|TInterface\\Icons\\spell_holy_wordfortitude"..ZOOMED_ICON.."|t",
    ["Holy"] = "|TInterface\\Icons\\spell_holy_guardianspirit"..ZOOMED_ICON.."|t",
    ["Shadow"] = "|TInterface\\Icons\\spell_shadow_shadowwordpain"..ZOOMED_ICON.."|t"
  },
  ["ROGUE"] = {
    ["Unknown"] = "|TInterface\\Icons\\INV_Misc_QuestionMark"..ZOOMED_ICON.."|t",
    ["Assassination"] = "|TInterface\\Icons\\ability_rogue_eviscerate"..ZOOMED_ICON.."|t",
    ["Combat"] = "|TInterface\\Icons\\ability_backstab"..ZOOMED_ICON.."|t",
    ["Subtlety"] = "|TInterface\\Icons\\ability_stealth"..ZOOMED_ICON.."|t"
  },
  ["SHAMAN"] = {
    ["Unknown"] = "|TInterface\\Icons\\INV_Misc_QuestionMark"..ZOOMED_ICON.."|t",
    ["Elemental"] = "|TInterface\\Icons\\spell_nature_lightning"..ZOOMED_ICON.."|t",
    ["Enhancement"] = "|TInterface\\Icons\\spell_nature_lightningshield"..ZOOMED_ICON.."|t",
    ["Restoration"] = "|TInterface\\Icons\\spell_nature_magicimmunity"..ZOOMED_ICON.."|t"
  },
  ["WARLOCK"] = {
    ["Unknown"] = "|TInterface\\Icons\\INV_Misc_QuestionMark"..ZOOMED_ICON.."|t",
    ["Affliction"] = "|TInterface\\Icons\\spell_shadow_deathcoil"..ZOOMED_ICON.."|t",
    ["Demonology"] = "|TInterface\\Icons\\spell_shadow_metamorphosis"..ZOOMED_ICON.."|t",
    ["Destruction"] = "|TInterface\\Icons\\spell_shadow_rainoffire"..ZOOMED_ICON.."|t"
  },
  ["WARRIOR"] = {
    ["Unknown"] = "|TInterface\\Icons\\INV_Misc_QuestionMark"..ZOOMED_ICON.."|t",
    ["Arms"] = "|TInterface\\Icons\\ability_rogue_eviscerate"..ZOOMED_ICON.."|t",
    ["Fury"] = "|TInterface\\Icons\\Ability_Warrior_InnerRage"..ZOOMED_ICON.."|t",
    ["Protection"] = "|TInterface\\Icons\\inv_shield_06"..ZOOMED_ICON.."|t"
  }
}

local possible_stuns = {
  ["Improved HoJ - 6s"] = {
    require = {
      class = "PALADIN",
      talent = "Improved Hammer of Justice"
    },
    icon = "|TInterface\\Icons\\spell_holy_sealofmight"..ZOOMED_ICON_SMALL.."|t",
    priority = 2,
    cooldown = 40,
    spellName = "Hammer of Justice"
  },
  ["HoJ - 6s"] = {
    require = {
      class = "PALADIN"
    },
    icon = "|TInterface\\Icons\\spell_holy_sealofmight"..ZOOMED_ICON_SMALL.."|t",
    priority = 1,
    cooldown = 60,
    spellName = "Hammer of Justice"
  },
  ["Maim - 6s"] = {
    require = {
      class = "DRUID",
      spec = "Feral Combat"
    },
    icon = "|TInterface\\Icons\\ability_druid_mangle"..ZOOMED_ICON_SMALL.."|t",
    priority = 1,
    cooldown = 0,
    spellName = "Maim"
  },
  ["Kidney Shot - 6s"] = {
    require = {
      class = "ROGUE"
    },
    icon = "|TInterface\\Icons\\ability_rogue_kidneyshot"..ZOOMED_ICON_SMALL.."|t",
    priority = 1,
    cooldown = 0,
    spellName = "Kidney Shot"
  },
  ["Concussion Blow - 5s"] = {
    require = {
      class = "WARRIOR",
      talent = "Concussion Blow"
    },
    icon = "|TInterface\\Icons\\ability_thunderbolt"..ZOOMED_ICON_SMALL.."|t",
    priority = 1,
    cooldown = 30,
    spellName = "Concussion Blow"
  },
  ["Deep Freeze - 5s"] = {
    require = {
      class = "MAGE",
      talent = "Deep Freeze"
    },
    icon = "|TInterface\\Icons\\ability_mage_deepfreeze"..ZOOMED_ICON_SMALL.."|t",
    priority = 1,
    cooldown = 30,
    spellName = "Deep Freeze"
  },
  ["Bash - 5s"] = {
    require = {
      class = "DRUID",
      talent = "Brutal Impact"
    },
    icon = "|TInterface\\Icons\\ability_druid_bash"..ZOOMED_ICON_SMALL.."|t",
    priority = 2,
    cooldown = 30,
    spellName = "Bash"
  },
  --[[["Bash - 4s"] = {
    class = "DRUID",
    talent = {
      ["Feral(Tank)"] = true
    },
    icon = "|TInterface\\Icons\\ability_druid_bash"..ZOOMED_ICON_SMALL.."|t",
    priority = 1,
    rotation = true
  },]]
  ["Gnaw - 3s"] = {
    require = {
      class = "DEATHKNIGHT",
      talent = "Master of Ghouls"
    },
    icon = "|TInterface\\Icons\\spell_deathknight_gnaw_ghoul"..ZOOMED_ICON_SMALL.."|t",
    rotation = true,
    priority = 1,
    cooldown = 60,
    spellName = "Gnaw"
  },
  ["Intimidation - 3s"] = {
    require = {
      class = "HUNTER",
      talent = "Intimidation"
    },
    icon = "|TInterface\\Icons\\ability_devour"..ZOOMED_ICON_SMALL.."|t",
    rotation = true,
    priority = 1,
    cooldown = 60,
    spellName = "Intimidation"
  }
}

local possible_slows = {
  ["Chains of Ice"] = {
    require = {
      class = "DEATHKNIGHT"
    },
    icon = "|TInterface\\Icons\\spell_frost_chainsofice"..ZOOMED_ICON_SMALL.."|t",
    priority = 1,
    spellName = "Chains of Ice"
  },
  ["Infected Wounds"] = {
    require = {
      class = "DRUID",
      talent = "Infected Wounds"
    },
    icon = "|TInterface\\Icons\\ability_druid_infectedwound"..ZOOMED_ICON_SMALL.."|t",
    priority = 1,
    spellName = "Infected Wounds"
  },
  ["Wing Clip"] = {
    require = {
      class = "HUNTER"
    },
    icon = "|TInterface\\Icons\\ability_rogue_trip"..ZOOMED_ICON_SMALL.."|t",
    priority = 1,
    spellName = "Wing Clip"
  },
  ["Frostbolt"] = {
    require = {
      class = "MAGE"
    },
    icon = "|TInterface\\Icons\\spell_frost_frostbolt02"..ZOOMED_ICON_SMALL.."|t",
    priority = 1,
    spellName = "Frostbolt"
  },
  ["Frostfire Bolt"] = {
    require = {
      class = "MAGE",
      spec = "Fire"
    },
    icon = "|TInterface\\Icons\\ability_mage_frostfirebolt"..ZOOMED_ICON_SMALL.."|t",
    priority = 2,
    spellName = "Frostfire Bolt"
  },
  ["Slow"] = {
    require = {
      class = "MAGE",
      talent = "Slow"
    },
    icon = "|TInterface\\Icons\\spell_nature_slow"..ZOOMED_ICON_SMALL.."|t",
    priority = 3,
    spellName = "Slow"
  },
  ["Mind Flay"] = {
    require = {
      class = "PRIEST",
      spec = "Shadow"
    },
    icon = "|TInterface\\Icons\\spell_shadow_siphonmana"..ZOOMED_ICON_SMALL.."|t",
    priority = 1,
    spellName = "Mind Flay"
  },
  ["Crippling Poison"] = {
    require = {
      class = "ROGUE"
    },
    icon = "|TInterface\\Icons\\ability_poisonsting"..ZOOMED_ICON_SMALL.."|t",
    priority = 1,
    spellName = "Crippling Poison"
  },
  ["Frost Shock"] = {
    require = {
      class = "SHAMAN"
    },
    icon = "|TInterface\\Icons\\spell_frost_frostshock"..ZOOMED_ICON_SMALL.."|t",
    priority = 1,
    spellName = "Frost Shock"
  },
  ["Curse of Exhaustion (30%)"] = {
    require = {
      class = "WARLOCK",
      talent = "Curse of Exhaustion"
    },
    icon = "|TInterface\\Icons\\spell_shadow_grimward"..ZOOMED_ICON_SMALL.."|t",
    priority = 1,
    spellName = "Curse of Exhaustion"
  },
  ["Hamstring"] = {
    require = {
      class = "WARRIOR"
    },
    icon = "|TInterface\\Icons\\ability_shockwave"..ZOOMED_ICON_SMALL.."|t",
    priority = 1,
    spellName = "Hamstring"
  }
}

local possible_aoe_stuns = {
  ["Shockwave - 4s"] = {
    require = {
      class = "WARRIOR",
      talent = "Shockwave"
    },
    icon = "|TInterface\\Icons\\ability_warrior_shockwave"..ZOOMED_ICON_SMALL.."|t",
    priority = 1,
    cooldown = 20,
    spellName = "Shockwave"
  },
  ["Holy Wrath - 3s"] = {
    require = {
      class = "PALADIN"
    },
    icon = "|TInterface\\Icons\\spell_holy_excorcism"..ZOOMED_ICON_SMALL.."|t",
    priority = 1,
    cooldown = 30,
    spellName = "Holy Wrath"
  },
  ["Shadowfury - 3s"] = {
    require = {
      class = "WARLOCK",
      talent = "Shadowfury"
    },
    icon = "|TInterface\\Icons\\spell_shadow_shadowfury"..ZOOMED_ICON_SMALL.."|t",
    priority = 1,
    cooldown = 20,
    spellName = "Shadowfury"
  }
}

local possible_aoe_slows = {
  --[[["Piercing Howl"] = {
    require = {
      class = "WARRIOR",
      talent = "Piercing Howl"
    },
    icon = "|TInterface\\Icons\\spell_shadow_deathscream"..ZOOMED_ICON_SMALL.."|t",
    priority = 1
  }, --Does Piercing Howl work on Val'kyr?]]
  ["Frost Trap"] = {
    require = {
      class = "HUNTER"
    },
    icon = "|TInterface\\Icons\\spell_frost_freezingbreath"..ZOOMED_ICON_SMALL.."|t",
    priority = 1,
    spellName = "Frost Trap"
  },
  ["Chilblains"] = {
    require = {
      class = "DEATHKNIGHT",
      talent = "Chilblains"
    },
    icon = "|TInterface\\Icons\\spell_frost_wisp"..ZOOMED_ICON_SMALL.."|t",
    priority = 1,
    spellName = "Pestilence"
  },
  ["Cone of Cold"] = {
    require = {
      class = "MAGE"
    },
    icon = "|TInterface\\Icons\\spell_frost_glacier"..ZOOMED_ICON_SMALL.."|t",
    priority = 1,
    spellName = "Cone of Cold"
  }
}

local possible_infests = {
  ["Aura Mastery - 2m"] = {
    require = {
      class = "PALADIN",
      talent = "Aura Mastery"
    },
    icon = "|TInterface\\Icons\\spell_holy_auramastery"..ZOOMED_ICON_SMALL.."|t",
    cooldown = 120,
    vehicle = true,
    priority = 0,
    spellName = "Aura Mastery"
  },
  ["Divine Sacrifice - 2m"] = {
    require = {
      class = "PALADIN",
      talent = "Divine Sacrifice"
    },
    icon = "|TInterface\\Icons\\spell_holy_powerwordbarrier"..ZOOMED_ICON_SMALL.."|t",
    cooldown = 120,
    priority = 0,
    spellName = "Divine Sacrifice"
  },
  ["Divine Hymn - 6m"] = {
    require = {
      class = "PRIEST"
    },
    icon = "|TInterface\\Icons\\spell_holy_divinehymn"..ZOOMED_ICON_SMALL.."|t",
    cooldown = 360,
    priority = 1,
    spellName = "Divine Hymn"
  }
}

local options = {
  type = "group",
  args = {}
}

local tooltip_update_frame = CreateFrame("FRAME");
local LKADB = LDB:NewDataObject("LKAnnounce", {
  type = "data source",
  text = "LKAnnounce",
  icon = "Interface\\Icons\\achievement_boss_lichking",
  --Interface\\Icons\\achievement_boss_svalasorrowgrave
  --Interface\\TARGETINGFRAME\\UI-TargetingFrame-Skull
	OnClick = function(self, button)
		if(IsShiftKeyDown()) then
			LKA_Announce_Stuns();
      LKA_Announce_Slows();
      LKA_AnnounceInfests();
		else
			InterfaceOptionsFrame_OpenToCategory("LKAnnounce");
		end
	end,
	OnEnter = function(self)
		local elapsed = 0;
		local delay = 1;
		tooltip_update_frame:SetScript("OnUpdate", function(self, elap)
			elapsed = elapsed + elap;
			if(elapsed > delay) then
				elapsed = 0;
				GameTooltip:ClearLines();
				GameTooltip:AddDoubleLine("LKAnnounce", loaded_header);
				GameTooltip:AddDoubleLine("By "..class_colors.MAGE.."Mirrored|r of Dragonmaw(US) Horde", "");
				GameTooltip:AddLine(" ");
				GameTooltip:AddLine("|cffeda55fClick|r for options", 0.2, 1, 0.2);
				GameTooltip:AddLine("|cffeda55fShift-Click|r to announce all assignments", 0.2, 1, 0.2);
				GameTooltip:Show();
			end
		end);
		
		--Section the screen into 6 quadrants (sextrants?) and define the tooltip anchor position based on which quadrant the cursor is in
		local max_x = GetScreenWidth();
		local max_y = GetScreenHeight();
		local x, y = GetCursorPosition();
		local horizontal = (x < (max_x/3) and "LEFT") or ((x >= (max_x/3) and x < ((max_x/3)*2)) and "") or "RIGHT";
		local tooltip_vertical = (y < (max_y/2) and "BOTTOM") or "TOP";
		local anchor_vertical = (y < (max_y/2) and "TOP") or "BOTTOM";

		GameTooltip:SetOwner(self, "ANCHOR_NONE");
		GameTooltip:SetPoint(tooltip_vertical..horizontal, self, anchor_vertical..horizontal);
		GameTooltip:AddDoubleLine("LKAnnounce", loaded_header);
		GameTooltip:AddDoubleLine("By "..class_colors.MAGE.."Mirrored|r of Dragonmaw(US) Horde", "");
		GameTooltip:AddLine(" ");
		GameTooltip:AddLine("|cffeda55fClick|r for options", 0.2, 1, 0.2);
		GameTooltip:AddLine("|cffeda55fShift-Click|r to announce all assignments", 0.2, 1, 0.2);
		GameTooltip:Show();
	end,
	OnLeave = function(self)
		tooltip_update_frame:SetScript("OnUpdate", nil);
	end
});

local alertTexture = UIParent:CreateTexture();

local chat_colors_frame = CreateFrame("FRAME");
chat_colors_frame:RegisterEvent("UPDATE_CHAT_COLOR");
chat_colors_frame:SetScript("OnEvent", function(self, event, ...)
  channel_colors.RAID = LKA_ratio_to_hex_color(ChatTypeInfo.RAID.r,ChatTypeInfo.RAID.g,ChatTypeInfo.RAID.b);
  channel_colors.RAID_WARNING = LKA_ratio_to_hex_color(ChatTypeInfo.RAID_WARNING.r, ChatTypeInfo.RAID_WARNING.g, ChatTypeInfo.RAID_WARNING.b);
  channel_colors.YELL = LKA_ratio_to_hex_color(ChatTypeInfo.YELL.r, ChatTypeInfo.YELL.g, ChatTypeInfo.YELL.b);
  channel_colors.PARTY = LKA_ratio_to_hex_color(ChatTypeInfo.PARTY.r, ChatTypeInfo.PARTY.g, ChatTypeInfo.PARTY.b);
  channel_colors.SAY = LKA_ratio_to_hex_color(ChatTypeInfo.SAY.r, ChatTypeInfo.SAY.g, ChatTypeInfo.SAY.b);
  channel_colors.GUILD = LKA_ratio_to_hex_color(ChatTypeInfo.GUILD.r, ChatTypeInfo.GUILD.g, ChatTypeInfo.GUILD.b);
  
  channels.RAID = "|cFF"..channel_colors.RAID.."Raid|r";
  channels.RAID_WARNING = "|cFF"..channel_colors.RAID_WARNING.."RW|r";
  channels.YELL = "|cFF"..channel_colors.YELL.."Yell|r";
  channels.PARTY = "|cFF"..channel_colors.PARTY.."Party|r";
  channels.SAY = "|cFF"..channel_colors.SAY.."Say|r";
  channels.GUILD = "|cFF"..channel_colors.GUILD.."Guild|r";
end);

local variables_loaded_frame = CreateFrame("FRAME");
variables_loaded_frame:RegisterEvent("VARIABLES_LOADED");
variables_loaded_frame:SetScript("OnEvent", function(self, event, ...)
  frame:RegisterEvent("PARTY_MEMBERS_CHANGED");
  
  if(LKAnnounceSaved.valks == nil) then
    LKA_wipe_valk_assignments();
  end
  if(LKAnnounceSaved.options == nil) then
    LKA_DefaultOptions()
  end
  
  options = {
    type = "group",
    args = {
      debug = {
        type = "execute",
        name = "Debug",
        hidden = true,
        desc = "Enable debug messages",
        func = function()
          debug = true;
          LKA_message("Debug messages enabled");
        end,
        guiHidden = true
      },
      nodebug = {
        type = "execute",
        name = "No Debug",
        desc = "Turn off debug messages",
        func = function()
          debug = false;
          LKA_message("Debug messages disabled");
        end,
        guiHidden = true
      },
      config = {
        type = "execute",
        name = "Configure",
        order = 0,
        desc = "Open the configuration dialog",
        func = function()
          InterfaceOptionsFrame_OpenToCategory("LKAnnounce"); --Opens the config frame
        end,
        guiHidden = true
      },
      roles = {
        type = "group",
        name = "Raid Member Talents",
        order = 10,
        childGroups = "tree",
        desc = "Roles and talent specs of members of the raid.",
        args = {
          loaded = {
            type = "header",
            name = function(info) return loaded_header end,
            order = 0
          },
          main = {
            type = "group",
            name = "   ---Main Raid---",
            inline = false;
            disabled = true,
            order = 5,
            args = {}
          },
          extra = {
            type = "group",
            name = "   ---Extras---",
            inline = false;
            disabled = true,
            order = function(info)
              local difficulty = GetRaidDifficulty();
              if(difficulty == 1) then
                return 25;
              elseif(difficulty == 2) then
                return 55;
              elseif(difficulty == 3) then
                return 25;
              elseif(difficulty == 4) then
                return 55;
              end
              return 0;
            end,
            args = {}
          }
        }
      },
      assignments = {
        type = "group",
        name = "Assignments",
        order = 20,
        childGroups = "tab",
        desc = "Assignments for stuns and slows for Val'Kyr Shadowguards in phase 2, Infest cooldowns, and healing throughout the fight.",
        args = {
          loaded = {
            type = "header",
            name = function(info) return loaded_header end,
            order = 0
          },
          stuns = {
            type = "group",
            name = "Stuns",
            order = 10,
            desc = "Single target stuns for phase 2 Val'kyr",
            args = {
              dsc = {
                type = "description",
                order = 0,
                name = "Stuns for Val'kyr Shadowguards for phase 2 of The Lich King"
              },
              help = {
                type = "execute",
                name = "Help",
                width = "half",
                confirm = function(info) return ASSIGNMENT_HELP end,
                order = 1,
                desc = "Get help with assignments",
                func = function() --[[doesn't actually do anything]] end
              },
              wipe = {
                type = "execute",
                name = "Wipe",
                width = "half",
                confirm = function(info) return "Are you sure you want to wipe all stun assignments?" end,
                order = 2,
                desc = "Delete all assignments",
                func = function()
                  for i,v in pairs(LKAnnounceSaved.valks.stun) do
                    LKAnnounceSaved.valks.stun[i] = nil;
                  end
                  for i,v in pairs(LKAnnounceSaved.valks.aoestun) do
                    LKAnnounceSaved.valks.aoestun[i] = nil;
                  end
                end
              },
              announce = {
                type = "execute",
                name = "Announce",
                order = 3,
                desc = "Announce stun assignments to the raid",
                func = function() LKA_Announce_Stuns(); end
              },
              stuns_header = {
                type = "header",
                name = "Main Stuns",
                order = 5,
              },
              stun1 = {
                type = "select",
                name = function(info) return shapes.content[LKAnnounceSaved.valks.shapes[1]].name end,
                width = "double",
                order = 10,
                desc = function(info) return "Main assigned stun for "..shapes.content[LKAnnounceSaved.valks.shapes[1]].name end,
                values = function(info) return stuns:names("stun", 1) end,
                get = function(info) return LKA_Get_Assignment("stun", 1) end,
                set = function(info, v) LKA_Set_Assignment("stun", 1, v) end
              },
              stun1b = {
                type = "select",
                name = "",
                width = "double",
                order = 15,
                hidden = function(info) return not rotation_stuns.content[LKAnnounceSaved.valks.stun[1]] end,
                desc = function(info) return "Second in stun rotation for "..shapes.content[LKAnnounceSaved.valks.shapes[1]].name end,
                values = function(info) return rotation_stuns:names("stun", 11) end,
                get = function(info) return LKA_Get_Assignment("stun", 11) end,
                set = function(info, v) LKA_Set_Assignment("stun", 11, v) end
              },
              stun2 = {
                type = "select",
                name = function(info) return shapes.content[LKAnnounceSaved.valks.shapes[2]].name end,
                width = "double",
                order = 20,
                desc = function(info) return "Main assigned stun for "..shapes.content[LKAnnounceSaved.valks.shapes[2]].name end,
                values = function(info) return stuns:names("stun", 2) end,
                get = function(info) return LKA_Get_Assignment("stun", 2) end,
                set = function(info, v) LKA_Set_Assignment("stun", 2, v) end
              },
              stun2b = {
                type = "select",
                name = "",
                width = "double",
                order = 25,
                hidden = function(info) return not rotation_stuns.content[LKAnnounceSaved.valks.stun[2]] end,
                desc = function(info) return "Second in stun rotation for "..shapes.content[LKAnnounceSaved.valks.shapes[2]].name end,
                values = function(info) return rotation_stuns:names("stun", 12) end,
                get = function(info) return LKA_Get_Assignment("stun", 12) end,
                set = function(info, v) LKA_Set_Assignment("stun", 12, v) end
              },
              stun3 = {
                type = "select",
                name = function(info) return shapes.content[LKAnnounceSaved.valks.shapes[3]].name end,
                width = "double",
                order = 30,
                desc = function(info) return "Main assigned stun for "..shapes.content[LKAnnounceSaved.valks.shapes[3]].name end,
                values = function(info) return stuns:names("stun", 3) end,
                get = function(info) return LKA_Get_Assignment("stun", 3) end,
                set = function(info, v) LKA_Set_Assignment("stun", 3, v) end
              },
              stun3b = {
                type = "select",
                name = "",
                width = "double",
                order = 35,
                hidden = function(info) return not rotation_stuns.content[LKAnnounceSaved.valks.stun[3]] end,
                desc = function(info) return "Second in stun rotation for "..shapes.content[LKAnnounceSaved.valks.shapes[3]].name end,
                values = function(info) return rotation_stuns:names("stun", 13) end,
                get = function(info) return LKA_Get_Assignment("stun", 13) end,
                set = function(info, v) LKA_Set_Assignment("stun", 13, v) end
              },
              backup_stuns_header = {
                type = "header",
                name = "Backup Stuns",
                order = 40,
              },
              backupstun1 = {
                type = "select",
                name = "Backup stun 1",
                width = "double",
                order = 50,
                desc = "First backup stun",
                values = function(info) return stuns:names("stun", 4) end,
                get = function(info) return LKA_Get_Assignment("stun", 4) end,
                set = function(info, v) LKA_Set_Assignment("stun", 4, v) end
              },
              backupstun2 = {
                type = "select",
                name = "Backup stun 2",
                width = "double",
                order = 60,
                desc = "Second backup stun",
                values = function(info) return stuns:names("stun", 5) end,
                get = function(info) return LKA_Get_Assignment("stun", 5) end,
                set = function(info, v) LKA_Set_Assignment("stun", 5, v) end
              },
              backupstun3 = {
                type = "select",
                name = "Backup stun 3",
                width = "double",
                order = 70,
                desc = "Third backup stun",
                values = function(info) return stuns:names("stun", 6) end,
                get = function(info) return LKA_Get_Assignment("stun", 6) end,
                set = function(info, v) LKA_Set_Assignment("stun", 6, v) end
              },
              dsc = {
                type = "header",
                order = 80,
                name = "AoE Stuns"
              },
              aoestun1 = {
                type = "select",
                name = "AoE stun 1",
                width = "double",
                order = 81,
                desc = "First AoE stun",
                values = function(info) return aoe_stuns:names("aoestun", 1) end,
                get = function(info) return LKA_Get_Assignment("aoestun", 1) end,
                set = function(info, v) LKA_Set_Assignment("aoestun", 1, v) end
              },
              aoestun2 = {
                type = "select",
                name = "AoE stun 2",
                width = "double",
                order = 82,
                desc = "Second AoE stun",
                values = function(info) return aoe_stuns:names("aoestun", 2) end,
                get = function(info) return LKA_Get_Assignment("aoestun", 2) end,
                set = function(info, v) LKA_Set_Assignment("aoestun", 2, v) end
              },
              aoestun3 = {
                type = "select",
                name = "AoE stun 3",
                width = "double",
                order = 83,
                desc = "Third AoE stun",
                values = function(info) return aoe_stuns:names("aoestun", 3) end,
                get = function(info) return LKA_Get_Assignment("aoestun", 3) end,
                set = function(info, v) LKA_Set_Assignment("aoestun", 3, v) end
              }
            }
          },
          slows = {
            type = "group",
            name = "Slows",
            order = 20,
            desc = "Single target slows for phase 2 Val'kyr",
            args = {
              dsc = {
                type = "description",
                order = 0,
                name = "Slows for Val'kyr Shadowguards for phase 2 of The Lich King"
              },
              help = {
                type = "execute",
                name = "Help",
                width = "half",
                confirm = function(info) return ASSIGNMENT_HELP end,
                order = 1,
                desc = "Get help with assignments",
                func = function() --[[doesn't actually do anything]] end
              },
              wipe = {
                type = "execute",
                name = "Wipe",
                width = "half",
                order = 2,
                confirm = function(info) return "Are you sure you want to wipe all stun assignments?" end,
                desc = "Delete all assignments",
                func = function()
                  for i,v in pairs(LKAnnounceSaved.valks.slow) do
                    LKAnnounceSaved.valks.slow[i] = nil;
                  end
                  for i,v in pairs(LKAnnounceSaved.valks.aoeslow) do
                    LKAnnounceSaved.valks.aoeslow[i] = nil;
                  end
                end
              },
              announce = {
                type = "execute",
                name = "Announce",
                order = 3,
                desc = "Announce slow assignments to the raid",
                func = function() LKA_Announce_Slows(); end
              },
              slows_header = {
                type = "header",
                name = "Main Slows",
                order = 10,
              },
              slow1 = {
                type = "select",
                name = function(info) return shapes.content[LKAnnounceSaved.valks.shapes[1]].name end,
                width = "double",
                order = 11,
                desc = function(info) return "Main assigned slow for "..shapes.content[LKAnnounceSaved.valks.shapes[1]].name end,
                values = function(info) return slows:names("slow", 1) end,
                get = function(info) return LKA_Get_Assignment("slow", 1) end,
                set = function(info, v) LKA_Set_Assignment("slow", 1, v) end
              },
              slow2 = {
                type = "select",
                name = function(info) return shapes.content[LKAnnounceSaved.valks.shapes[2]].name end,
                width = "double",
                order = 12,
                desc = function(info) return "Main assigned slow for "..shapes.content[LKAnnounceSaved.valks.shapes[2]].name end,
                values = function(info) return slows:names("slow", 2) end,
                get = function(info) return LKA_Get_Assignment("slow", 2) end,
                set = function(info, v) LKA_Set_Assignment("slow", 2, v) end
              },
              slow3 = {
                type = "select",
                name = function(info) return shapes.content[LKAnnounceSaved.valks.shapes[3]].name end,
                width = "double",
                order = 13,
                desc = function(info) return "Main assigned slow for "..shapes.content[LKAnnounceSaved.valks.shapes[3]].name end,
                values = function(info) return slows:names("slow", 3) end,
                get = function(info) return LKA_Get_Assignment("slow", 3) end,
                set = function(info, v) LKA_Set_Assignment("slow", 3, v) end
              },
              backup_slows_header = {
                type = "header",
                name = "Backup Slows",
                order = 20,
              },
              backupslow1 = {
                type = "select",
                name = "Backup slow 1",
                width = "double",
                order = 21,
                desc = "First backup slow",
                values = function(info) return slows:names("slow", 4) end,
                get = function(info) return LKA_Get_Assignment("slow", 4) end,
                set = function(info, v) LKA_Set_Assignment("slow", 4, v) end
              },
              backupslow2 = {
                type = "select",
                name = "Backup slow 2",
                width = "double",
                order = 22,
                desc = "Second backup slow",
                values = function(info) return slows:names("slow", 5) end,
                get = function(info) return LKA_Get_Assignment("slow", 5) end,
                set = function(info, v) LKA_Set_Assignment("slow", 5, v) end
              },
              backupslow3 = {
                type = "select",
                name = "Backup slow 3",
                width = "double",
                order = 23,
                desc = "Third backup slow",
                values = function(info) return slows:names("slow", 6) end,
                get = function(info) return LKA_Get_Assignment("slow", 6) end,
                set = function(info, v) LKA_Set_Assignment("slow", 6, v) end
              },
              dsc = {
                type = "header",
                order = 30,
                name = "AoE Slows"
              },
              aoeslow1 = {
                type = "select",
                name = "AoE slow 1",
                width = "double",
                order = 31,
                desc = "First AoE slow",
                values = function(info) return aoe_slows:names("aoeslow", 1) end,
                get = function(info) return LKA_Get_Assignment("aoeslow", 1) end,
                set = function(info, v) LKA_Set_Assignment("aoeslow", 1, v) end
              },
              aoeslow2 = {
                type = "select",
                name = "AoE slow 2",
                width = "double",
                order = 32,
                desc = "Second AoE slow",
                values = function(info) return aoe_slows:names("aoeslow", 2) end,
                get = function(info) return LKA_Get_Assignment("aoeslow", 2) end,
                set = function(info, v) LKA_Set_Assignment("aoeslow", 2, v) end
              },
              aoeslow3 = {
                type = "select",
                name = "AoE slow 3",
                width = "double",
                order = 33,
                desc = "Third AoE slow",
                values = function(info) return aoe_slows:names("aoeslow", 3) end,
                get = function(info) return LKA_Get_Assignment("aoeslow", 3) end,
                set = function(info, v) LKA_Set_Assignment("aoeslow", 3, v) end
              }
            }
          },
          infests = {
            type = "group",
            name = "Infests",
            order = 30,
            desc = "Infests cooldown assignments for phases 1 and 2",
            args = {
              add = {
                type = "execute",
                order = 1,
                name = "+",
                width = "half",
                desc = "Add an infest assignment",
                func = function() LKA_AddInfest() end
              },
              remove = {
                type = "execute",
                order = 2,
                name = "-",
                width = "half",
                disabled = true,
                desc = "Add an infest assignment",
                func = function() LKA_RemoveInfest() end
              },
              announce = {
                type = "execute",
                order = 3,
                name = "Announce",
                width = "normal",
                desc = "Announce infest assignments",
                func = function() LKA_AnnounceInfests() end
              }
            }
          },
          heals = {
            type = "group",
            name = "Healing",
            order = 60,
            desc = "Healing assignments for phase 2",
            args = {
              name = {
                type = "input",
                order = 10,
                name = "Add Healer",
                get = function(info) end,
                set = function(info, v) LKA_AddHealer(v); end
              },
              announce = {
                type = "execute",
                order = 20,
                name = "Announce",
                func = function(info) LKA_AnnounceHealingAssignments() end
              }
            }
          }
        }
      },
      options = {
        type = "group",
        name = "Options",
        order = 30,
        childGroups = "tree",
        desc = "Options for announcements throughout the fight.",
        args = {
          loaded = {
            type = "header",
            name = function(info) return loaded_header end,
            order = 0
          },
          plague = {
            type = "group",
            name = "Plagues & Horrors",
            order = 10,
            childGroups = "tree",
            desc = "Options for Necrotic Plague and Shambling Horror announcements",
            args = {
              track_header = {
                type = "header",
                order = 0,
                name = "Track Plagues"
              },
              track_desc = {
                type = "description",
                order = 2,
                name = "Announce estimated frenzy and death times of Shambling Horrors due to Necrotic Plague."
              },
              track_toggle = {
                type = "toggle",
                name = "Enable",
                width = "half",
                order = 4,
                get = function(info) return LKAnnounceSaved.options.plague.track.enabled end,
                set = function(info, v) LKAnnounceSaved.options.plague.track.enabled = v end
              },
              track_channel = {
                type = "select",
                name = "Channel",
                width = "half",
                order = 6,
                values = channels,
                disabled = function(info) return not LKAnnounceSaved.options.plague.track.enabled end,
                get = function(info) return LKAnnounceSaved.options.plague.track.channel end,
                set = function(info, v) LKAnnounceSaved.options.plague.track.channel = v end
              },
              enrage_header = {
                type = "header",
                order = 8,
                name = "Enrage"
              },
              enrage_desc = {
                type = "description",
                order = 10,
                name = "Announce when Shambling Horrors cast Enrage."
              },
              enrage_toggle = {
                type = "toggle",
                name = "Enable",
                width = "half",
                order = 12,
                get = function(info) return LKAnnounceSaved.options.plague.enrage.enabled end,
                set = function(info, v) LKAnnounceSaved.options.plague.enrage.enabled = v end
              },
              enrage_channel = {
                type = "select",
                name = "Channel",
                width = "half",
                order = 14,
                values = channels,
                disabled = function(info) return not LKAnnounceSaved.options.plague.enrage.enabled end,
                get = function(info) return LKAnnounceSaved.options.plague.enrage.channel end,
                set = function(info, v) LKAnnounceSaved.options.plague.enrage.channel = v end
              },
              frenzy_header = {
                type = "header",
                order = 16,
                name = "Frenzy"
              },
              frenzy_desc = {
                type = "description",
                order = 18,
                name = "Announce when Shambling Horrors cast Frenzy (heroic only)."
              },
              frenzy_toggle = {
                type = "toggle",
                name = "Enable",
                width = "half",
                order = 20,
                get = function(info) return LKAnnounceSaved.options.plague.frenzy.enabled end,
                set = function(info, v) LKAnnounceSaved.options.plague.frenzy.enabled = v end
              },
              frenzy_channel = {
                type = "select",
                name = "Channel",
                width = "half",
                order = 22,
                values = channels,
                disabled = function(info) return not LKAnnounceSaved.options.plague.frenzy.enabled end,
                get = function(info) return LKAnnounceSaved.options.plague.frenzy.channel end,
                set = function(info, v) LKAnnounceSaved.options.plague.frenzy.channel = v end
              }
            }
          },
          infest = {
            type = "group",
            name = "Infests",
            order = 20,
            childGroups = "tree",
            desc = "Options for Infest announcements",
            args = {
              infest_header = {
                type = "header",
                order = 0,
                name = "Numbered Infests"
              },
              infest_desc = {
                type = "description",
                order = 2,
                name = "Number infests and preemptively announce when Infests are about to occur."
              },
              infest_toggle1 = {
                type = "toggle",
                name = "Phase 1",
                width = "half",
                order = 4,
                get = function(info) return LKAnnounceSaved.options.infest.number.enabled1 end,
                set = function(info, v) LKAnnounceSaved.options.infest.number.enabled1 = v end
              },
              infest_channel1 = {
                type = "select",
                name = "Channel",
                width = "half",
                order = 6,
                values = channels,
                disabled = function(info) return not LKAnnounceSaved.options.infest.number.enabled1 end,
                get = function(info) return LKAnnounceSaved.options.infest.number.channel1 end,
                set = function(info, v) LKAnnounceSaved.options.infest.number.channel1 = v end
              },
              infest_toggle2 = {
                type = "toggle",
                name = "Phase 2",
                width = "half",
                order = 8,
                get = function(info) return LKAnnounceSaved.options.infest.number.enabled2 end,
                set = function(info, v) LKAnnounceSaved.options.infest.number.enabled2 = v end
              },
              infest_channel2 = {
                type = "select",
                name = "Channel",
                width = "half",
                order = 10,
                values = channels,
                disabled = function(info) return not LKAnnounceSaved.options.infest.number.enabled2 end,
                get = function(info) return LKAnnounceSaved.options.infest.number.channel2 end,
                set = function(info, v) LKAnnounceSaved.options.infest.number.channel2 = v end
              },
              cooldown_header = {
                type = "header",
                order = 12,
                name = "Infest Cooldowns"
              },
              cooldown_desc = {
                type = "description",
                order = 14,
                name = "Add cooldown suggestions to the Infest announcements (see the Assignments tab)."
              },
              cooldown_toggle1 = {
                type = "toggle",
                name = "Phase 1",
                width = "half",
                order = 16,
                disabled = function(info) return not LKAnnounceSaved.options.infest.number.enabled1 end,
                get = function(info) return LKAnnounceSaved.options.infest.cooldown.enabled1 end,
                set = function(info, v) LKAnnounceSaved.options.infest.cooldown.enabled1 = v end
              },
              cooldown_channel1 = {
                type = "select",
                name = " ",
                width = "half",
                order = 18,
                values = channels,
                disabled = true,
                get = function(info) return LKAnnounceSaved.options.infest.number.channel1 end,
                set = function(info, v) LKAnnounceSaved.options.infest.number.channel1 = v end
              },
              cooldown_toggle2 = {
                type = "toggle",
                name = "Phase 2",
                width = "half",
                order = 20,
                disabled = function(info) return not LKAnnounceSaved.options.infest.number.enabled2 end,
                get = function(info) return LKAnnounceSaved.options.infest.cooldown.enabled2 end,
                set = function(info, v) LKAnnounceSaved.options.infest.cooldown.enabled2 = v end
              },
              cooldown_channel2 = {
                type = "select",
                name = " ",
                width = "half",
                order = 22,
                values = channels,
                disabled = true,
                get = function(info) return LKAnnounceSaved.options.infest.number.channel2 end,
                set = function(info, v) LKAnnounceSaved.options.infest.number.channel2 = v end
              }
            }
          },
          valkyr = {
            type = "group",
            name = "Val'kyr Shadowguards",
            order = 30,
            childGroups = "tree",
            desc = "Options for Val'kyr Shadowguard announcements",
            args = {
              picked_header = {
                type = "header",
                name = "Val'kyr Picks",
                order = 0
              },
              picked_desc = {
                type = "description",
                name = "Announce who is picked by Val'kyr Shadowguards",
                order = 1
              },
              picked_toggle = {
                type = "toggle",
                name = "Enable",
                width = "half",
                order = 2,
                get = function(info) return LKAnnounceSaved.options.valkyr.picked.enabled end,
                set = function(info, v) LKAnnounceSaved.options.valkyr.picked.enabled = v end
              },
              picked_channel = {
                type = "select",
                name = "Channel",
                width = "half",
                order = 3,
                values = channels,
                disabled = function(info) return not LKAnnounceSaved.options.valkyr.picked.enabled end,
                get = function(info) return LKAnnounceSaved.options.valkyr.picked.channel end,
                set = function(info, v) LKAnnounceSaved.options.valkyr.picked.channel = v end
              },
              marks_header = {
                type = "header",
                name = "Marks",
                order = 4
              },
              marks = {
                type = "select",
                name = "Shapes",
                order = 5,
                values = function(info) return shapes_sets:names() end,
                get = function(info)
                  local set = nil;
                  for i,v in pairs(shapes_sets.content) do
                    if(v.marks[1] == LKAnnounceSaved.valks.shapes[1] and v.marks[2] == LKAnnounceSaved.valks.shapes[2] and v.marks[3] == LKAnnounceSaved.valks.shapes[3]) then
                      set = i;
                      break;
                    end
                  end
                  return set;
                end,
                set = function(info, v)
                  local set = shapes_sets.content[v];
                  LKAnnounceSaved.valks.shapes[1] = set.marks[1];
                  LKAnnounceSaved.valks.shapes[2] = set.marks[2];
                  LKAnnounceSaved.valks.shapes[3] = set.marks[3];
                end
              },
              marks_toggle = {
                type = "toggle",
                name = "Mark Val'kyr",
                order = 6,
                --descStyle = "inline",
                desc = "|cFFFF0000Not Yet Implemented|r",
                disabled = function(info)
                  if(LKAnnounceSaved.valks.shapes[1] + LKAnnounceSaved.valks.shapes[2] + LKAnnounceSaved.valks.shapes[3] == 6) then
                    return false;
                  end
                  return true;
                end,
                get = function(info)
                  if(LKAnnounceSaved.valks.shapes[1] + LKAnnounceSaved.valks.shapes[2] + LKAnnounceSaved.valks.shapes[3] == 6) then
                    return LKAnnounceSaved.options.valkyr.picked.mark;
                  end
                  return false;
                end,
                set = function(info, v) LKAnnounceSaved.options.valkyr.picked.mark = v end
              },
              stun = {
                type = "group",
                name = "Stuns",
                order = 8,
                args = {
                  toggle = {
                    type = "toggle",
                    name = "Enable",
                    width = "full",
                    order = 0,
                    get = function(info) return LKAnnounceSaved.options.valkyr.stun.enabled end,
                    set = function(info, v) LKAnnounceSaved.options.valkyr.stun.enabled = v end
                  },
                  backup_header = {
                    type = "header",
                    name = "Backups",
                    order = 2,
                    hidden = function(info) return not LKAnnounceSaved.options.valkyr.stun.enabled end
                  },
                  backup_desc = {
                    type = "description",
                    name = "Announce Val'kyr stun backups when the main assigned raiders are disabled.",
                    order = 4,
                    hidden = function(info) return not LKAnnounceSaved.options.valkyr.stun.enabled end
                  },
                  backup_toggle = {
                    type = "toggle",
                    name = "Enable",
                    width = "half",
                    order = 6,
                    hidden = function(info) return not LKAnnounceSaved.options.valkyr.stun.enabled end,
                    get = function(info) return LKAnnounceSaved.options.valkyr.stun.backup_enabled end,
                    set = function(info, v) LKAnnounceSaved.options.valkyr.stun.backup_enabled = v end
                  },
                  backup_channel = {
                    type = "select",
                    name = "Channel",
                    width = "half",
                    order = 8,
                    values = channels,
                    hidden = function(info) return not LKAnnounceSaved.options.valkyr.stun.enabled end,
                    disabled = function(info) return not LKAnnounceSaved.options.valkyr.stun.backup_enabled end,
                    get = function(info) return LKAnnounceSaved.options.valkyr.stun.backup_channel end,
                    set = function(info, v) LKAnnounceSaved.options.valkyr.stun.backup_channel = v end
                  },
                  backup_whisper = {
                    type = "toggle",
                    name = "Whisper Assignments",
                    width = "full",
                    order = 10,
                    hidden = function(info) return not LKAnnounceSaved.options.valkyr.stun.enabled end,
                    disabled = function(info) return not LKAnnounceSaved.options.valkyr.stun.backup_enabled end,
                    get = function(info) return LKAnnounceSaved.options.valkyr.stun.backup_whisper end,
                    set = function(info, v) LKAnnounceSaved.options.valkyr.stun.backup_whisper = v end
                  },
                  main_header = {
                    type = "header",
                    name = "Main",
                    order = 12,
                    hidden = function(info) return not LKAnnounceSaved.options.valkyr.stun.enabled end
                  },
                  main_desc = {
                    type = "description",
                    name = "Announce Val'kyr stuns even when they do not differ from the main assignment.",
                    order = 14,
                    hidden = function(info) return not LKAnnounceSaved.options.valkyr.stun.enabled end
                  },
                  main_toggle = {
                    type = "toggle",
                    name = "Enable",
                    width = "half",
                    order = 16,
                    hidden = function(info) return not LKAnnounceSaved.options.valkyr.stun.enabled end,
                    get = function(info) return LKAnnounceSaved.options.valkyr.stun.main_enabled end,
                    set = function(info, v) LKAnnounceSaved.options.valkyr.stun.main_enabled = v end
                  },
                  main_channel = {
                    type = "select",
                    name = "Channel",
                    width = "half",
                    order = 18,
                    values = channels,
                    hidden = function(info) return not LKAnnounceSaved.options.valkyr.stun.enabled end,
                    get = function(info) return LKAnnounceSaved.options.valkyr.stun.main_channel end,
                    set = function(info, v) LKAnnounceSaved.options.valkyr.stun.main_channel = v end
                  },
                  main_whisper = {
                    type = "toggle",
                    name = "Whisper Assignments",
                    width = "full",
                    order = 20,
                    hidden = function(info) return not LKAnnounceSaved.options.valkyr.stun.enabled end,
                    disabled = function(info) return not LKAnnounceSaved.options.valkyr.stun.main_enabled end,
                    get = function(info) return LKAnnounceSaved.options.valkyr.stun.main_whisper end,
                    set = function(info, v) LKAnnounceSaved.options.valkyr.stun.main_whisper = v end
                  }
                }
              },
              aoestun = {
                type = "group",
                name = "AoE Stuns",
                order = 10,
                args = {
                  toggle = {
                    type = "toggle",
                    name = "Enable",
                    width = "full",
                    order = 0,
                    get = function(info) return LKAnnounceSaved.options.valkyr.aoestun.enabled end,
                    set = function(info, v) LKAnnounceSaved.options.valkyr.aoestun.enabled = v end
                  },
                  backup_header = {
                    type = "header",
                    name = "Backups",
                    order = 2,
                    hidden = function(info) return not LKAnnounceSaved.options.valkyr.aoestun.enabled end
                  },
                  backup_desc = {
                    type = "description",
                    name = "Announce a Val'kyr AoE stun backup when the main assigned raider is disabled.",
                    order = 4,
                    hidden = function(info) return not LKAnnounceSaved.options.valkyr.aoestun.enabled end
                  },
                  backup_toggle = {
                    type = "toggle",
                    name = "Enable",
                    width = "half",
                    order = 6,
                    hidden = function(info) return not LKAnnounceSaved.options.valkyr.aoestun.enabled end,
                    get = function(info) return LKAnnounceSaved.options.valkyr.aoestun.backup_enabled end,
                    set = function(info, v) LKAnnounceSaved.options.valkyr.aoestun.backup_enabled = v end
                  },
                  backup_channel = {
                    type = "select",
                    name = "Channel",
                    width = "half",
                    order = 8,
                    values = channels,
                    hidden = function(info) return not LKAnnounceSaved.options.valkyr.aoestun.enabled end,
                    disabled = function(info) return not LKAnnounceSaved.options.valkyr.aoestun.backup_enabled end,
                    get = function(info) return LKAnnounceSaved.options.valkyr.aoestun.backup_channel end,
                    set = function(info, v) LKAnnounceSaved.options.valkyr.aoestun.backup_channel = v end
                  },
                  backup_whisper = {
                    type = "toggle",
                    name = "Whisper Assignments",
                    width = "full",
                    order = 10,
                    hidden = function(info) return not LKAnnounceSaved.options.valkyr.aoestun.enabled end,
                    get = function(info) return LKAnnounceSaved.options.valkyr.aoestun.backup_whisper end,
                    set = function(info, v) LKAnnounceSaved.options.valkyr.aoestun.backup_whisper = v end
                  },
                  main_header = {
                    type = "header",
                    name = "Main",
                    order = 12,
                    hidden = function(info) return not LKAnnounceSaved.options.valkyr.aoestun.enabled end
                  },
                  main_desc = {
                    type = "description",
                    name = "Announce the Val'kyr AoE stun even when it does not differ from the main assignment.",
                    order = 14,
                    hidden = function(info) return not LKAnnounceSaved.options.valkyr.aoestun.enabled end
                  },
                  main_toggle = {
                    type = "toggle",
                    name = "Enable",
                    width = "half",
                    order = 16,
                    hidden = function(info) return not LKAnnounceSaved.options.valkyr.aoestun.enabled end,
                    get = function(info) return LKAnnounceSaved.options.valkyr.aoestun.main_enabled end,
                    set = function(info, v) LKAnnounceSaved.options.valkyr.aoestun.main_enabled = v end
                  },
                  main_channel = {
                    type = "select",
                    name = "Channel",
                    width = "half",
                    order = 18,
                    values = channels,
                    hidden = function(info) return not LKAnnounceSaved.options.valkyr.aoestun.enabled end,
                    disabled = function(info) return not LKAnnounceSaved.options.valkyr.aoestun.main_enabled end,
                    get = function(info) return LKAnnounceSaved.options.valkyr.aoestun.main_channel end,
                    set = function(info, v) LKAnnounceSaved.options.valkyr.aoestun.main_channel = v end
                  },
                  main_whisper = {
                    type = "toggle",
                    name = "Whisper Assignments",
                    width = "full",
                    order = 20,
                    hidden = function(info) return not LKAnnounceSaved.options.valkyr.aoestun.enabled end,
                    get = function(info) return LKAnnounceSaved.options.valkyr.aoestun.main_whisper end,
                    set = function(info, v) LKAnnounceSaved.options.valkyr.aoestun.main_whisper = v end
                  }
                }
              },
              slow = {
                type = "group",
                name = "Slows",
                order = 12,
                args = {
                  toggle = {
                    type = "toggle",
                    name = "Enable",
                    width = "full",
                    order = 0,
                    get = function(info) return LKAnnounceSaved.options.valkyr.slow.enabled end,
                    set = function(info, v) LKAnnounceSaved.options.valkyr.slow.enabled = v end
                  },
                  backup_header = {
                    type = "header",
                    name = "Backups",
                    order = 2,
                    hidden = function(info) return not LKAnnounceSaved.options.valkyr.slow.enabled end
                  },
                  backup_desc = {
                    type = "description",
                    name = "Announce Val'kyr slow backups when the main assigned raiders are disabled.",
                    order = 4,
                    hidden = function(info) return not LKAnnounceSaved.options.valkyr.slow.enabled end
                  },
                  backup_toggle = {
                    type = "toggle",
                    name = "Enable",
                    width = "half",
                    order = 6,
                    hidden = function(info) return not LKAnnounceSaved.options.valkyr.slow.enabled end,
                    get = function(info) return LKAnnounceSaved.options.valkyr.slow.backup_enabled end,
                    set = function(info, v) LKAnnounceSaved.options.valkyr.slow.backup_enabled = v end
                  },
                  backup_channel = {
                    type = "select",
                    name = "Channel",
                    width = "half",
                    order = 8,
                    values = channels,
                    hidden = function(info) return not LKAnnounceSaved.options.valkyr.slow.enabled end,
                    disabled = function(info) return not LKAnnounceSaved.options.valkyr.slow.backup_enabled end,
                    get = function(info) return LKAnnounceSaved.options.valkyr.slow.backup_channel end,
                    set = function(info, v) LKAnnounceSaved.options.valkyr.slow.backup_channel = v end
                  },
                  backup_whisper = {
                    type = "toggle",
                    name = "Whisper Assignments",
                    width = "full",
                    order = 10,
                    hidden = function(info) return not LKAnnounceSaved.options.valkyr.slow.enabled end,
                    get = function(info) return LKAnnounceSaved.options.valkyr.slow.backup_whisper end,
                    set = function(info, v) LKAnnounceSaved.options.valkyr.slow.backup_whisper = v end
                  },
                  main_header = {
                    type = "header",
                    name = "Main",
                    order = 12,
                    hidden = function(info) return not LKAnnounceSaved.options.valkyr.slow.enabled end
                  },
                  main_desc = {
                    type = "description",
                    name = "Announce Val'kyr slows even when they do not differ from the main assignment.",
                    order = 14,
                    hidden = function(info) return not LKAnnounceSaved.options.valkyr.slow.enabled end
                  },
                  main_toggle = {
                    type = "toggle",
                    name = "Enable",
                    width = "half",
                    order = 16,
                    hidden = function(info) return not LKAnnounceSaved.options.valkyr.slow.enabled end,
                    get = function(info) return LKAnnounceSaved.options.valkyr.slow.main_enabled end,
                    set = function(info, v) LKAnnounceSaved.options.valkyr.slow.main_enabled = v end
                  },
                  main_channel = {
                    type = "select",
                    name = "Channel",
                    width = "half",
                    order = 18,
                    values = channels,
                    hidden = function(info) return not LKAnnounceSaved.options.valkyr.slow.enabled end,
                    disabled = function(info) return not LKAnnounceSaved.options.valkyr.slow.main_enabled end,
                    get = function(info) return LKAnnounceSaved.options.valkyr.slow.main_channel end,
                    set = function(info, v) LKAnnounceSaved.options.valkyr.slow.main_channel = v end
                  },
                  main_whisper = {
                    type = "toggle",
                    name = "Whisper Assignments",
                    width = "full",
                    order = 20,
                    hidden = function(info) return not LKAnnounceSaved.options.valkyr.slow.enabled end,
                    get = function(info) return LKAnnounceSaved.options.valkyr.slow.main_whisper end,
                    set = function(info, v) LKAnnounceSaved.options.valkyr.slow.main_whisper = v end
                  }
                }
              },
              aoeslow = {
                type = "group",
                name = "AoE Slows",
                order = 14,
                args = {
                  toggle = {
                    type = "toggle",
                    name = "Enable",
                    width = "full",
                    order = 0,
                    get = function(info) return LKAnnounceSaved.options.valkyr.aoeslow.enabled end,
                    set = function(info, v) LKAnnounceSaved.options.valkyr.aoeslow.enabled = v end
                  },
                  backup_header = {
                    type = "header",
                    name = "Backups",
                    order = 2,
                    hidden = function(info) return not LKAnnounceSaved.options.valkyr.aoeslow.enabled end
                  },
                  backup_desc = {
                    type = "description",
                    name = "Announce a Val'kyr AoE stun backup when the main assigned raider is disabled.",
                    order = 4,
                    hidden = function(info) return not LKAnnounceSaved.options.valkyr.aoeslow.enabled end
                  },
                  backup_toggle = {
                    type = "toggle",
                    name = "Enable",
                    width = "half",
                    order = 6,
                    hidden = function(info) return not LKAnnounceSaved.options.valkyr.aoeslow.enabled end,
                    get = function(info) return LKAnnounceSaved.options.valkyr.aoeslow.backup_enabled end,
                    set = function(info, v) LKAnnounceSaved.options.valkyr.aoeslow.backup_enabled = v end
                  },
                  backup_channel = {
                    type = "select",
                    name = "Channel",
                    width = "half",
                    order = 8,
                    values = channels,
                    hidden = function(info) return not LKAnnounceSaved.options.valkyr.aoeslow.enabled end,
                    disabled = function(info) return not LKAnnounceSaved.options.valkyr.aoeslow.backup_enabled end,
                    get = function(info) return LKAnnounceSaved.options.valkyr.aoeslow.backup_channel end,
                    set = function(info, v) LKAnnounceSaved.options.valkyr.aoeslow.backup_channel = v end
                  },
                  backup_whisper = {
                    type = "toggle",
                    name = "Whisper Assignments",
                    width = "full",
                    order = 10,
                    hidden = function(info) return not LKAnnounceSaved.options.valkyr.aoeslow.enabled end,
                    get = function(info) return LKAnnounceSaved.options.valkyr.aoeslow.backup_whisper end,
                    set = function(info, v) LKAnnounceSaved.options.valkyr.aoeslow.backup_whisper = v end
                  },
                  main_header = {
                    type = "header",
                    name = "Main",
                    order = 12,
                    hidden = function(info) return not LKAnnounceSaved.options.valkyr.aoeslow.enabled end
                  },
                  main_desc = {
                    type = "description",
                    name = "Announce the Val'kyr AoE stun even when it does not differ from the main assignment.",
                    order = 14,
                    hidden = function(info) return not LKAnnounceSaved.options.valkyr.aoeslow.enabled end
                  },
                  main_toggle = {
                    type = "toggle",
                    name = "Enable",
                    width = "half",
                    order = 16,
                    hidden = function(info) return not LKAnnounceSaved.options.valkyr.aoeslow.enabled end,
                    get = function(info) return LKAnnounceSaved.options.valkyr.aoeslow.main_enabled end,
                    set = function(info, v) LKAnnounceSaved.options.valkyr.aoeslow.main_enabled = v end
                  },
                  main_channel = {
                    type = "select",
                    name = "Channel",
                    width = "half",
                    order = 18,
                    values = channels,
                    hidden = function(info) return not LKAnnounceSaved.options.valkyr.aoeslow.enabled end,
                    disabled = function(info) return not LKAnnounceSaved.options.valkyr.aoeslow.main_enabled end,
                    get = function(info) return LKAnnounceSaved.options.valkyr.aoeslow.main_channel end,
                    set = function(info, v) LKAnnounceSaved.options.valkyr.aoeslow.main_channel = v end
                  },
                  main_whisper = {
                    type = "toggle",
                    name = "Whisper Assignments",
                    width = "full",
                    order = 20,
                    hidden = function(info) return not LKAnnounceSaved.options.valkyr.aoeslow.enabled end,
                    get = function(info) return LKAnnounceSaved.options.valkyr.aoeslow.main_whisper end,
                    set = function(info, v) LKAnnounceSaved.options.valkyr.aoeslow.main_whisper = v end
                  }
                }
              }
            }
          },
          defile = {
            type = "group",
            name = "Defile",
            order = 40,
            childGroups = "tree",
            desc = "Options for Defile announcements",
            args = {
              warning1 = {
                type = "description",
                order = 0,
                name = "|cFFFF0000Warning: |rDefile announcements usually do not work unless you put The Lich King on /focus."
              },
              warning2 = {
                type = "description",
                order = 1,
                name = "|cFFFF0000Warning: |rDefile announcements do not register for the main tank (because The Lich King does not change targets)."
              },
              announce = {
                type = "group",
                name = "Announce",
                order = 10,
                args = {
                  announce_header = {
                    type = "header",
                    order = 0,
                    name = "Announce"
                  },
                  announce_desc = {
                    type = "description",
                    order = 2,
                    name = "Announce Defile targets."
                  },
                  announce_toggle2 = {
                    type = "toggle",
                    name = "Phase 2",
                    width = "half",
                    order = 4,
                    get = function(info) return LKAnnounceSaved.options.defile.announce.enabled2 end,
                    set = function(info, v) LKAnnounceSaved.options.defile.announce.enabled2 = v end
                  },
                  announce_channel2 = {
                    type = "select",
                    name = "Channel",
                    width = "half",
                    order = 6,
                    values = channels,
                    disabled = function(info) return not LKAnnounceSaved.options.defile.announce.enabled2 end,
                    get = function(info) return LKAnnounceSaved.options.defile.announce.channel2 end,
                    set = function(info, v) LKAnnounceSaved.options.defile.announce.channel2 = v end
                  },
                  announce_toggle3 = {
                    type = "toggle",
                    name = "Phase 3",
                    width = "half",
                    order = 8,
                    get = function(info) return LKAnnounceSaved.options.defile.announce.enabled3 end,
                    set = function(info, v) LKAnnounceSaved.options.defile.announce.enabled3 = v end
                  },
                  announce_channel3 = {
                    type = "select",
                    name = "Channel",
                    width = "half",
                    order = 10,
                    values = channels,
                    disabled = function(info) return not LKAnnounceSaved.options.defile.announce.enabled3 end,
                    get = function(info) return LKAnnounceSaved.options.defile.announce.channel3 end,
                    set = function(info, v) LKAnnounceSaved.options.defile.announce.channel3 = v end
                  },
                  announce_role = {
                    type = "toggle",
                    name = "Add Target's Role",
                    width = "normal",
                    order = 12,
                    get = function(info) return LKAnnounceSaved.options.defile.announce.role end,
                    set = function(info, v) LKAnnounceSaved.options.defile.announce.role = v end
                  },
                  announce_whisper = {
                    type = "toggle",
                    name = "Whisper Target",
                    width = "normal",
                    order = 14,
                    get = function(info) return LKAnnounceSaved.options.defile.announce.whisper end,
                    set = function(info, v) LKAnnounceSaved.options.defile.announce.whisper = v end
                  }
                }
              },
              alert = {
                type = "group",
                name = "Alert",
                order = 15,
                args = {
                  alert_header = {
                    type = "header",
                    order = 16,
                    name = "Alert"
                  },
                  alert_desc = {
                    type = "description",
                    order = 18,
                    name = "Flash your screen a certain color when defile is casting."
                  },
                  alert_self_toggle = {
                    type = "toggle",
                    name = "On you",
                    width = "normal",
                    order = 20,
                    get = function(info) return LKAnnounceSaved.options.defile.alert.self end,
                    set = function(info, v) LKAnnounceSaved.options.defile.alert.self = v end
                  },
                  alert_self_color = {
                    type = "color",
                    name = "Color",
                    width = "half",
                    order = 22,
                    disabled = function(info) return not LKAnnounceSaved.options.defile.alert.self end,
                    hasAlpha = true,
                    get = function(info)
                      local c = LKAnnounceSaved.options.defile.alert.self_color;
                      return c.r, c.g, c.b, c.a;
                    end,
                    set = function(info, r, g, b, a)
                      local c = LKAnnounceSaved.options.defile.alert.self_color;
                      c.r, c.g, c.b, c.a = r, g, b, a;
                    end
                  },
                  alert_self_test = {
                    type = "execute",
                    name = "Test",
                    width = "half",
                    order = 24,
                    disabled = function(info) return not LKAnnounceSaved.options.defile.alert.self end,
                    func = function(info)
                      local c = LKAnnounceSaved.options.defile.alert.self_color;
                      LKA_FlashScreen(c.r, c.g, c.b, c.a);
                    end
                  },
                  alert_other_toggle = {
                    type = "toggle",
                    name = "On others",
                    width = "normal",
                    order = 26,
                    get = function(info) return LKAnnounceSaved.options.defile.alert.other end,
                    set = function(info, v) LKAnnounceSaved.options.defile.alert.other = v end
                  },
                  alert_other_color = {
                    type = "color",
                    name = "Color",
                    width = "half",
                    order = 28,
                    disabled = function(info) return not LKAnnounceSaved.options.defile.alert.other end,
                    hasAlpha = true,
                    get = function(info)
                      c = LKAnnounceSaved.options.defile.alert.other_color;
                      return c.r, c.g, c.b, c.a;
                    end,
                    set = function(info, r, g, b, a)
                      c = LKAnnounceSaved.options.defile.alert.other_color;
                      c.r, c.g, c.b, c.a = r, g, b, a;
                    end
                  },
                  alert_other_test = {
                    type = "execute",
                    name = "Test",
                    width = "half",
                    order = 30,
                    disabled = function(info) return not LKAnnounceSaved.options.defile.alert.other end,
                    func = function(info)
                      local c = LKAnnounceSaved.options.defile.alert.other_color;
                      LKA_FlashScreen(c.r, c.g, c.b, c.a);
                    end
                  }
                }
              }
            }
          },
          heal = {
            type = "group",
            name = "Healing",
            order = 50,
            childGroups = "tree",
            desc = "Options for healing assignment announcements throughout the fight",
            args = {
              disabled_header = {
                type = "header",
                name = "Disabled Announcements",
                order = 0
              },
              disabled_desc = {
                type = "description",
                name = "Announce a healer's assignment if he/she is disabled (see the Assignments tab for healing assignments)",
                order = 2
              },
              disabled_channel = {
                type = "select",
                name = "Channel",
                width = "half",
                order = 3,
                values = channels,
                disabled = function(info)
                  return not (LKAnnounceSaved.options.heal.disabled.death or LKAnnounceSaved.options.heal.disabled.disconnect or LKAnnounceSaved.options.heal.disabled.leave_raid or LKAnnounceSaved.options.heal.disabled.valkyr or LKAnnounceSaved.options.heal.disabled.harvest_soul);
                end,
                get = function(info) return LKAnnounceSaved.options.heal.disabled.channel end,
                set = function(info, v) LKAnnounceSaved.options.heal.disabled.channel = v end
              },
              disabled_announce = {
                type = "multiselect",
                name = "Announce on:",
                width = "double",
                order = 4,
                values = LKAnnounceSaved.options.heal.disabled.options_list,
                get = function(info, v)
                  return LKAnnounceSaved.options.heal.disabled[v];
                end,
                set = function(info, v)
                  LKAnnounceSaved.options.heal.disabled[v] = not LKAnnounceSaved.options.heal.disabled[v];
                end
              }
            }
          },
          trap = {
            type = "group",
            name = "Traps & Spheres",
            order = 55,
            desc = "Options for Traps and Spheres",
            args = {
              desc = {
                type = "description",
                order = 10,
                name = "Make a whimsical announcement when someone hits a Shadow Trap or Ice Sphere"
              },
              enable = {
                type = "toggle",
                name = "Enable",
                width = "half",
                order = 20,
                get = function(info) return LKAnnounceSaved.options.trap.fly.enabled end,
                set = function(info, v) LKAnnounceSaved.options.trap.fly.enabled = v end
              },
              channel = {
                type = "select",
                name = "Channel",
                width = "half",
                order = 30,
                values = channels,
                disabled = function(info) return not LKAnnounceSaved.options.trap.fly.enabled end,
                get = function(info) return LKAnnounceSaved.options.trap.fly.channel end,
                set = function(info, v) LKAnnounceSaved.options.trap.fly.channel = v end
              }
            }
          },
          cooldowns = {
            type = "group",
            name = "Cooldowns",
            order = 60,
            childGroups = "tree",
            desc = "Options for external cooldown announcements",
            args = {
              nyi = {
                type = "header",
                name = "|cFFFF0000Not Yet Implemented|r"
              }
            }
          }
        }
      }
    }
  };
  
  for i,v in pairs(LKAnnounceSaved.valks.heal) do
    LKA_AddHealer(i);
  end

  for i,v in pairs(LKAnnounceSaved.valks.infest) do
    LKA_AddInfest();
  end
  
  LibStub("AceConfig-3.0"):RegisterOptionsTable("LKAnnounce", options, {"lkannounce", "lka"});
  aboutFrame = LibStub("LibAboutPanel").new(nil, "LKAnnounce");
  LibStub("AceConfigDialog-3.0"):AddToBlizOptions("LKAnnounce", "Talents Loaded", "LKAnnounce", "roles");
  LibStub("AceConfigDialog-3.0"):AddToBlizOptions("LKAnnounce", "Assignments", "LKAnnounce", "assignments");
  LibStub("AceConfigDialog-3.0"):AddToBlizOptions("LKAnnounce", "Options", "LKAnnounce", "options");
  
  talentLib.RegisterCallback("LKAnnounce", "LibGroupTalents_Update", function(self, GUID, UID, spec, t1, t2, t3)
    LKA_UpdateHeaders();
    LibStub("AceConfigRegistry-3.0"):NotifyChange("LKAnnounce");
  end);
	
	modelFrame = CreateFrame("FRAME", nil, aboutFrame);
	modelFrame:SetWidth(240);
	modelFrame:SetHeight(240);
	modelFrame:SetPoint("BOTTOM", aboutFrame, "BOTTOM", 0, 10);
	modelFrame:SetBackdrop({
		bgFile = "Interface/Tooltips/UI-Tooltip-Background", 
    edgeFile = "Interface/Tooltips/UI-Tooltip-Border", 
    tile = true, tileSize = 16, edgeSize = 16, 
		insets = {
			left = 4,
			right = 4,
			top = 4,
			bottom = 4
		}
	});
	modelFrame:SetBackdropColor(0, 0, 0, 0.8);

	model = CreateFrame("PLAYERMODEL", nil, modelFrame);
  local model_time = 0;
	modelFrame:SetScript("OnShow", function()
		model:SetPoint("BOTTOMLEFT", modelFrame, "BOTTOMLEFT", 4, 4);
		model:SetPoint("TOPRIGHT", modelFrame, "TOPRIGHT", -4, -4);
		model:SetModel("Creature/Arthaslichking/arthaslichking.m2");
		--model:SetPosition(0, 0, 0);
		--model:SetLight(1, 0, 0, -0.707, -0.707, 0.7, 1.0, 1.0, 1.0, 0.8, 1.0, 1.0, 0.8);
	end);
	modelFrame:SetScript("OnUpdate", function(self, elapsed)
    model_time = model_time + (elapsed * 1000)
    if(model_time > 5000) then
      model_time = 0;
    end
		model:SetSequenceTime(84, model_time)
	end);
end);

function LKA_FlashScreen(r, g, b, a)
	alertTexture:SetTexture(r, g, b, a);
	alertTexture:SetAllPoints(UIParent);
	alertTexture:Show();

	Chronos.schedule(0.5, function()
		alertTexture:Hide();
	end);
end

function LKA_Get_Assignment(assign, index)
  if(LKAnnounceSaved.valks[assign][index]) then
    return LKAnnounceSaved.valks[assign][index];
  else
    return NONE;
  end
end

function LKA_Set_Assignment(assign, index, value)
  if(value == NONE) then
    LKAnnounceSaved.valks[assign][index] = nil;
  else
    LKAnnounceSaved.valks[assign][index] = value;
    for i,v in pairs(LKAnnounceSaved.valks[assign]) do
      if(i ~= index and v == value) then
        LKAnnounceSaved.valks[assign][i] = nil;
      end
    end
  end
end

function LKA_RemoveRotationSelect(index)
  local lengthenIndex = "stun"..index;
  local name = "stun"..index.."b";
  
  options.args.assignments.args.stuns.args[lengthenIndex].width = "double";
  options.args.assignments.args.stuns.args[name] = nil;
end

function LKA_AddInfest()
  if(num_infests < 9) then
    num_infests = num_infests + 1;
    local num = num_infests;
    
    options.args.assignments.args.infests.args["infest"..num] = {
      type = "select",
      name = "Infest "..num,
      order = num_infests * 10,
      desc = "#"..num.." in the rotation for cooldowns during infests",
      values = function(info) return infests:names("infest", num) end,
      width = "double",
      get = function(info) return LKA_Get_Assignment("infest", num, infests_list) end,
      set = function(info, v) LKA_Set_Assignment("infest", num, v) end
    }
    if(num_infests >= 1) then
      options.args.assignments.args.infests.args.remove.disabled = false;
    end
    if(num_infests >= 9) then
      options.args.assignments.args.infests.args.add.disabled = true;
    end
  end
end

function LKA_RemoveInfest()
  if(num_infests > 0) then
    local num = num_infests;
    options.args.assignments.args.infests.args["infest"..num] = nil;
    LKAnnounceSaved.valks.infest[num] = nil;
    num_infests = num_infests - 1;
    if(num_infests < 1) then
      options.args.assignments.args.infests.args.remove.disabled = true;
    end
    if(num_infests < 9) then
      options.args.assignments.args.infests.args.add.disabled = false;
    end
  end
end

function LKA_AnnounceInfests()
  if not((LKAnnounceSaved.options.infest.number.enabled1 and LKAnnounceSaved.options.infest.cooldown.enabled1) or (LKAnnounceSaved.options.infest.number.enabled2 and LKAnnounceSaved.options.infest.cooldown.enabled2)) then
    LKA_message("Infest cooldown announcements are enabled for neither Phase 1 nor Phase 2 - these assignments will never be announced mid-fight");
  end
  local num = 0;
  for i,v in ipairs(LKAnnounceSaved.valks.infest) do
    if(infests[v]) then
      num = num + 1;
      break;
    end
  end
  if(num > 0) then
    LKA_printraid("Infest Cooldowns:");
    local num = 0;
    for i,v in ipairs(LKAnnounceSaved.valks.infest) do
      if(infests[v]) then
        LKA_printraid("  "..i..": "..v);
        num = num + 1;
      end
    end
  else
    LKA_message("No Infest cooldowns to announce (old assignments are not announced)");
  end
end

function LKA_AnnounceHealingAssignments()
  local num = 0;
  for i,v in pairs(LKAnnounceSaved.valks.heal) do
    if(UnitInRaid(i)) then
      num = num + 1;
      break;
    end
  end
  if(num > 0) then
    for i,v in pairs(LKAnnounceSaved.valks.heal) do
      if(UnitInRaid(i)) then
        SendChatMessage(i..": "..v);
      end
    end
  else
    LKA_message("No healing assignments to announce (people not in raid are not announced)");
  end
end
      

function LKA_AddHealer(name)
  if(name and name ~= "") then
    local coloredName = name;
    local _, class = UnitClass(name);
    if(class) then
      coloredName = class_colors[class]..name.."|r";
    elseif(LKAnnounceSaved.roles[name]) then
      coloredName = class_colors[LKAnnounceSaved.roles[name].class]..name.."|r";
    end
    
    options.args.assignments.args.heals.args[name] = {
      type = "group",
      childGroups = "tree",
      name = coloredName,
      args = {
        assignment = {
          type = "input",
          order = 10,
          name = "Assignment",
          multiline = true,
          width = "full",
          get = function(info) return LKAnnounceSaved.valks.heal[info[#info-1]] end,
          set = function(info, v) LKAnnounceSaved.valks.heal[info[#info-1]] = v end
        },
        remove = {
          type = "execute",
          order = 20,
          name = "Remove",
          width = "normal",
          func = function(info) LKA_RemoveHealer(info[#info-1]) end
        }
      }
    };
  end
end

function LKA_RemoveHealer(name)
  LKAnnounceSaved.valks.heal[name] = nil;
  options.args.assignments.args.heals.args[name] = nil;
end

function LKA_Announce_Stuns()
  if(LKAnnounceSaved.options.valkyr.stun.enabled) then
    local mains = "Stuns: ";
    for i=1,3 do
      if(stuns.content[LKAnnounceSaved.valks.stun[i]]) then
        if(rotation_stuns.content[LKAnnounceSaved.valks.stun[i]] and rotation_stuns.content[LKAnnounceSaved.valks.stun[i + 10]]) then
          mains = mains..shapes.content[LKAnnounceSaved.valks.shapes[i]].chatName.." "..rotation_stuns.content[LKAnnounceSaved.valks.stun[i]].playerName.." and "..rotation_stuns.content[LKAnnounceSaved.valks.stun[i + 10]].playerName..", ";
        else
          mains = mains..shapes.content[LKAnnounceSaved.valks.shapes[i]].chatName.." "..stuns.content[LKAnnounceSaved.valks.stun[i]].playerName..", ";
        end
      end
    end
    if(mains ~= "Stuns: ") then
      mains = string.sub(mains, 0, -3) --Cut off the trailing comma
      LKA_printraid(mains);
    else
      --If there were no stun assignments
      LKA_message("No stuns to announce (old assignments are not announced)");
    end
    
    local backups = "Backups: ";
    for i=4,6 do
      if(stuns.content[LKAnnounceSaved.valks.stun[i]]) then
        backups = backups..stuns.content[LKAnnounceSaved.valks.stun[i]].playerName..", ";
      end
    end
    if(backups ~= "Backups: ") then
      backups = string.sub(backups, 0, -3) --Cut off the trailing comma
      LKA_printraid(backups);
    else
      --If there were no backup stun assignments
      LKA_message("No backup stuns to announce (old assignments are not announced)");
    end
  else
    LKA_message("Stuns module is disabled");
  end
  
  if(LKAnnounceSaved.options.valkyr.aoestun.enabled) then
    local aoe = "AoE Stuns: ";
    for i=1,3 do
      if(aoe_stuns.content[LKAnnounceSaved.valks.aoestun[i]]) then
        aoe = aoe..aoe_stuns.content[LKAnnounceSaved.valks.aoestun[i]].playerName..", ";
      end
    end
    if(aoe ~= "AoE Stuns: ") then
      mains = string.sub(aoe, 0, -3) --Cut off the trailing comma
      LKA_printraid(aoe);
    else
      --If there were no aoe stun assignments
      LKA_message("No AoE stuns to announce (old assignments are not announced)");
    end
  else
    LKA_message("AoE Stuns module is disabled");
  end
end

function LKA_Announce_Slows()
  if(LKAnnounceSaved.options.valkyr.slow.enabled) then
    local mains = "Slows: ";
    for i=1,3 do
      if(slows.content[LKAnnounceSaved.valks.slow[i]]) then
        mains = mains..shapes.content[LKAnnounceSaved.valks.shapes[i]].chatName.." "..slows.content[LKAnnounceSaved.valks.slow[i]].playerName..", ";
      end
    end
    if(mains ~= "Slows: ") then
      --If there were no stun assignments, don't announce them at all
      mains = string.sub(mains, 0, -3) --Cut off the trailing comma
      LKA_printraid(mains);
    else
      --If there were no slow assignments
      LKA_message("No slows to announce (old assignments are not announced)");
    end
    
    local backups = "Backups: ";
    for i=4,6 do
      if(slows.content[LKAnnounceSaved.valks.slow[i]]) then
        backups = backups..slows.content[LKAnnounceSaved.valks.slow[i]].playerName..", ";
      end
    end
    if(backups ~= "Backups: ") then
      --If there were no backup assignments, don't announce them at all
      backups = string.sub(backups, 0, -3) --Cut off the trailing comma
      LKA_printraid(backups);
    else
      --If there were no back slow assignments
      LKA_message("No backup slows to announce (old assignments are not announced)");
    end
  else
    LKA_message("Slows module is disabled");
  end
  
  if(LKAnnounceSaved.options.valkyr.aoeslow.enabled) then
    local aoe = "AoE Slows: ";
    for i=1,3 do
      if(aoe_stuns.content[LKAnnounceSaved.valks.aoeslow[i]]) then
        aoe = aoe..aoe_stuns.content[LKAnnounceSaved.valks.aoeslow[i]].playerName..", ";
      end
    end
    if(aoe ~= "AoE Stuns: ") then
      --If there were no aoe slow assignments, don't announce them at all
      aoe = string.sub(aoe, 0, -3) --Cut off the trailing comma
      LKA_printraid(aoe);
    else
      --If there were no aoe slow assignments
      LKA_message("No AoE slows to announce (old assignments are not announced)");
    end
  else
    LKA_message("AoE Slows module is disabled");
  end
end

function LKA_OnLoad()
  frame = CreateFrame("FRAME", "LKAMessageFrame");
  frame:RegisterEvent("RAID_ROSTER_UPDATE");
	frame:RegisterEvent("PLAYER_REGEN_DISABLED");
	frame:RegisterEvent("PLAYER_REGEN_ENABLED");
  frame:SetScript("OnEvent", LKA_OnEvent);
  
  LKA_message("Loaded!");
end

function LKA_OnEvent(self, event, ...)
  if(event == "COMBAT_LOG_EVENT_UNFILTERED") then
    LKA_HandleCombatMessage(self, event, ...);
  elseif(event == "UNIT_TARGET") then
    local unit = select(1,...);
    local target = unit.."target";
    if(GetUnitName(target)) then
      if(GetUnitName(target) == "Shambling Horror") then
        --If someone changes their target to a horror, put them in the table
        targeting_horrors[unit] = UnitGUID(target);
        --If this horror doesn't have a name, give it a name
        if not(horror_names[UnitGUID(target)]) then
          local name = "Shambling Horror "..current_horror;
          current_horror = current_horror + 1;
          horror_names[UnitGUID(target)] = name;
        end
        --Check horrors for plagues, in case a horror with a plague is being targetted for the first time
        LKA_Calculate_Plagues();
      else
        --If someone changes their target to something that is not a horror, make sure they aren't in the table anymore - their previous target could have been a horror
        if(targeting_horrors[unit]) then
          targeting_horrors[unit] = nil;
        end
      end
    else
      --If someone drops their target, make sure they aren't in the table anymore - their previous target could have been a horror
      if(targeting_horrors[unit]) then
        targeting_horrors[unit] = nil;
      end
    end
    if(GetUnitName(unit) == "The Lich King") then
      if(pending_defile) then
        pending_defile = false;
        LKA_AnnounceDefile(GetUnitName(target));
      end
    end
  elseif(event == "PLAYER_REGEN_DISABLED") then
    if(GetZoneText() == ZONE and GetSubZoneText() == SUBZONE) then
      LKA_StartEncounter();
    end
  elseif(event == "PLAYER_REGEN_ENABLED") then
    LKA_EndEncounter();
  elseif(event == "RAID_ROSTER_UPDATE") then
    local changed = false;
    --Delete options of persons who are no longer in the raid
    for i,v in pairs(options.args.roles.args) do
      if (i ~= "loaded" and i ~= "main" and i ~= "extra" and not UnitInRaid(i)) then
        LKA_LeftRaid(i);
        changed = true;
      end
    end
    for i=1,GetNumRaidMembers() do
      local name, _, subgroup = GetRaidRosterInfo(i);

      raid_positions[name] = subgroup * 10;
      if not(options.args.roles.args[name]) then
        LKA_JoinedRaid(name);
        if(name) then
        else
          LKA_debug("Trying to add nobody("..name..")");
        end
        changed = true;
      end
    end
    if(changed) then
      LibStub("AceConfigRegistry-3.0"):NotifyChange("LKAnnounce");
    end
  elseif(event == "CHAT_MSG_MONSTER_YELL") then
    msg = select(1,...);
    if(msg == LKPULL) then
      LKA_message("Put Lich King on your focus!");
    end
  end
end

function LKA_StartEncounter()
  infest_count = 1;
	phase = 1;
  current_horror = 1;
  wipe(horror_names);
  
  LKA_message("Lich King encounter started!");
  encounter_enabled = true;
  frame:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED");
  frame:RegisterEvent("UNIT_TARGET");
end

function LKA_EndEncounter()
  if(encounter_enabled) then
    LKA_message("Lich King encounter ended.");
    encounter_enabled = false;
    frame:UnregisterEvent("COMBAT_LOG_EVENT_UNFILTERED");
    frame:UnregisterEvent("UNIT_TARGET");
  end
end
  
function LKA_JoinedRaid(name)
  LKA_AddRole(name);
  LKA_AddToAssignments(name);
  LKA_UpdateHeaders();
end

function LKA_LeftRaid(name)
  if(encounter_enabled and LKAnnounceSaved.options.heal.disabled.leave_raid and LKAnnouncedSaved.valks.heal[name]) then
    LKA_AnnounceHealingAssignment(name, "left the raid");
  end
  LKA_RemoveRole(name);
  LKA_RemoveFromAssignments(name);
  LKA_UpdateHeaders();
end

function LKA_UpdateHeaders()
  local num_raid = GetNumRaidMembers();
  if(num_raid > 0) then
    local num_loaded = 0;
    local UID;
    for i=1,num_raid do
      UID = "raid"..i;
      if(talentLib:GetUnitTalentSpec(UID)) then
        num_loaded = num_loaded + 1;
      end
    end
    
    local g = (num_loaded/num_raid) * 2;
    local r = 2 - g;
    g = (g > 1 and 1) or g;
    r = (r > 1 and 1) or r;
    local color = LKA_ratio_to_hex_color(r, g, 0);
    loaded_header = "Loaded: |cFF"..color..num_loaded.."/"..num_raid.."|r";
  else
    loaded_header = NORAID;
  end
end

function LKA_AddRole(name)
  if(name) then
    _, class = UnitClass(name);
    if(class) then
      local colored_name = class_colors[class]..name.."|r";
    else
      colored_name = name;
    end
    options.args.roles.args[name] = {
      type = "group",
      name = function(info)
        local name = info[#info];
        local coloredName = name;
        local _, class = UnitClass(name);
        if(class) then
          coloredName = class_colors[class]..coloredName.."|r";
          local spec = talentLib:GetUnitTalentSpec(name);
          if(spec) then
            coloredName = possible_roles[class][spec]..coloredName
          end
        end
        return coloredName;
      end,
			desc = "",
			descStyle = "inline",
      order = function(info) return raid_positions[info[#info]]; end,
      args = {
        header = {
          type = "header",
          name = function(info)
            local spec, t1, t2, t3 = talentLib:GetUnitTalentSpec(name);
            if(spec) then
              _, class = UnitClass(name);
              return class_colors[class]..spec.." ("..(t1 or "0").."/"..(t2 or "0").."/"..(t3 or "0")..")|r";
            end
            return "Not Loaded";
          end,
          order = 0
        }
      }
    };
    
    local stuns_list;
    for i,v in pairs(possible_stuns) do
      if(v.require.class == class) then
        if not(stuns_list) then
          stuns_list = {};
        end
        stuns_list[v.spellName] = v.icon.." "..v.spellName;
      end
    end
    if(stuns_list) then
      options.args.roles.args[name].args.stuns = {
        type = "multiselect",
        name = "Stuns",
        values = stuns_list,
        get = function(info, v) return stuns.content[info[#info-1].." - "..v] end,
        set = function(info, v) end,
        disabled = function(info) return not talentLib:GetUnitTalentSpec(name) end,
        order = 10
      };
    end
    
    local aoe_stuns_list;
    for i,v in pairs(possible_aoe_stuns) do
      if(v.require.class == class) then
        if not(aoe_stuns_list) then
          aoe_stuns_list = {};
        end
        aoe_stuns_list[v.spellName] = v.icon.." "..v.spellName;
      end
    end
    if(aoe_stuns_list) then
      options.args.roles.args[name].args.aoe_stuns = {
        type = "multiselect",
        name = "AoE Stuns",
        values = aoe_stuns_list,
        get = function(info, v) return aoe_stuns.content[info[#info-1].." - "..v] end,
        set = function(info, v) end,
        disabled = function(info) return not talentLib:GetUnitTalentSpec(name) end,
        order = 20
      };
    end
    
    local slows_list;
    for i,v in pairs(possible_slows) do
      if(v.require.class == class) then
        if not(slows_list) then
          slows_list = {};
        end
        slows_list[v.spellName] = v.icon.." "..v.spellName;
      end
    end
    if(slows_list) then
      options.args.roles.args[name].args.slows = {
        type = "multiselect",
        name = "Slows",
        values = slows_list,
        get = function(info, v) return slows.content[info[#info-1].." - "..v] end,
        set = function(info, v) end,
        disabled = function(info) return not talentLib:GetUnitTalentSpec(name) end,
        order = 30
      };
    end
    
    local aoe_slows_list;
    for i,v in pairs(possible_aoe_slows) do
      if(v.require.class == class) then
        if not(aoe_slows_list) then
          aoe_slows_list = {};
        end
        aoe_slows_list[v.spellName] = v.icon.." "..v.spellName;
      end
    end
    if(aoe_slows_list) then
      options.args.roles.args[name].args.aoe_slows = {
        type = "multiselect",
        name = "AoE Slows",
        values = aoe_slows_list,
        get = function(info, v) return aoe_slows.content[info[#info-1].." - "..v] end,
        set = function(info, v) end,
        disabled = function(info) return not talentLib:GetUnitTalentSpec(name) end,
        order = 40
      };
    end
    
    local infests_list;
    for i,v in pairs(possible_infests) do
      if(v.require.class == class) then
        if not(infests_list) then
          infests_list = {};
        end
        infests_list[v.spellName] = v.icon.." "..v.spellName;
      end
    end
    if(infests_list) then
      options.args.roles.args[name].args.infests = {
        type = "multiselect",
        name = "Infest Cooldowns",
        values = infests_list,
        get = function(info, v) return infests.content[info[#info-1].." - "..v] end,
        set = function(info, v) end,
        disabled = function(info) return not talentLib:GetUnitTalentSpec(name) end,
        order = 50
      };
    end
  end
end

function LKA_RemoveRole(name)
  options.args.roles.args[name] = nil
end
  
function LKA_AddToAssignments(name)
  LKA_RemoveFromAssignments(name);
  if(name and UnitClass(name)) then
    local _, class = UnitClass(name);
    
    for i1,v1 in pairs(possible_stuns) do
      local priority = 0;
      local qualify = true;
      for i2,v2 in pairs(v1.require) do
        if(i2 == "class" and qualify) then
          if(class ~= v2) then qualify = false end
        elseif(i2 == "talent" and qualify) then
          local talent = talentLib:UnitHasTalent(name, v2);
          if not(talent) then
            qualify = false;
          end
        elseif(i2 == "spec" and qualify) then
          local spec = talentLib:GetUnitTalentSpec(name);
          if not(spec == v2) then
            qualify = false;
          end
        end
      end
      if(qualify and (v1.priority >= priority)) then
        priority = v1.priority;
        if(v1.cooldown > 40) then
          rotation_stuns.content[name.." - "..v1.spellName] = {
            name = class_colors[class]..name.."|r:  "..v1.icon.." "..i1,
            inUseName = "*"..class_colors[class]..name.."|r:  "..v1.icon.." "..i1,
            playerName = name,
            cooldown = v1.cooldown,
            spellName = v1.spellName
          }
        end
        stuns.content[name.." - "..v1.spellName] = {
          name = class_colors[class]..name.."|r:  "..v1.icon.." "..i1,
          inUseName = "*"..class_colors[class]..name.."|r:  "..v1.icon.." "..i1,
          playerName = name,
          cooldown = v1.cooldown,
          spellName = v1.spellName
        }
      end
    end
    
    for i1,v1 in pairs(possible_aoe_stuns) do
      local priority = 0;
      local qualify = true;
      for i2,v2 in pairs(v1.require) do
        if(i2 == "class" and qualify) then
          if(class ~= v2) then qualify = false end
        elseif(i2 == "talent" and qualify) then
          local talent = talentLib:UnitHasTalent(name, v2);
          if not(talent) then
            qualify = false;
          end
        elseif(i2 == "spec" and qualify) then
          local spec = talentLib:GetUnitTalentSpec(name);
          if not(spec == v2) then
            qualify = false;
          end
        end
      end
      if(qualify and (v1.priority >= priority)) then
        priority = v1.priority;
        LKA_debug(name);
        aoe_stuns.content[name.." - "..v1.spellName] = {
          name = class_colors[class]..name.."|r:  "..v1.icon.." "..i1,
          inUseName = "*"..class_colors[class]..name.."|r:  "..v1.icon.." "..i1,
          playerName = name,
          cooldown = v1.cooldown,
          spellName = v1.spellName
        }
      end
    end
    
    for i1,v1 in pairs(possible_slows) do
      local priority = 0;
      local qualify = true;
      for i2,v2 in pairs(v1.require) do
        if(i2 == "class" and qualify) then
          if(class ~= v2) then qualify = false end
        elseif(i2 == "talent" and qualify) then
          local talent = talentLib:UnitHasTalent(name, v2);
          if not(talent) then
            qualify = false;
          end
        elseif(i2 == "spec" and qualify) then
          local spec = talentLib:GetUnitTalentSpec(name);
          if not(spec == v2) then
            qualify = false;
          end
        end
      end
      if(qualify and (v1.priority >= priority)) then
        priority = v1.priority;
        slows.content[name.." - "..v1.spellName] = {
          name = class_colors[class]..name.."|r:  "..v1.icon.." "..i1,
          inUseName = "*"..class_colors[class]..name.."|r:  "..v1.icon.." "..i1,
          playerName = name,
          cooldown = v1.cooldown,
          spellName = v1.spellName
        }
      end
    end
    
    for i1,v1 in pairs(possible_aoe_slows) do
      local priority = 0;
      local qualify = true;
      for i2,v2 in pairs(v1.require) do
        if(i2 == "class" and qualify) then
          if(class ~= v2) then qualify = false end
        elseif(i2 == "talent" and qualify) then
          local talent = talentLib:UnitHasTalent(name, v2);
          if not(talent) then
            qualify = false;
          end
        elseif(i2 == "spec" and qualify) then
          local spec = talentLib:GetUnitTalentSpec(name);
          if not(spec == v2) then
            qualify = false;
          end
        end
      end
      if(qualify and (v1.priority >= priority)) then
        priority = v1.priority;
        aoe_slows.content[name.." - "..v1.spellName] = {
          name = class_colors[class]..name.."|r:  "..v1.icon.." "..i1,
          inUseName = "*"..class_colors[class]..name.."|r:  "..v1.icon.." "..i1,
          playerName = name,
          cooldown = v1.cooldown,
          spellName = v1.spellName
        }
      end
    end
    
    for i1,v1 in pairs(possible_infests) do
      local priority = 0;
      local qualify = true;
      for i2,v2 in pairs(v1.require) do
        if(i2 == "class" and qualify) then
          if(class ~= v2) then qualify = false end
        elseif(i2 == "talent" and qualify) then
          local talent = talentLib:UnitHasTalent(name, v2);
          if not(talent) then
            qualify = false;
          end
        elseif(i2 == "spec" and qualify) then
          local spec = talentLib:GetUnitTalentSpec(name);
          if not(spec == v2) then
            qualify = false;
          end
        end
      end
      if(qualify and (v1.priority >= priority)) then
        priority = v1.priority;
        infests.content[name.." - "..v1.spellName] = {
          name = class_colors[class]..name.."|r:  "..v1.icon.." "..i1,
          inUseName = "*"..class_colors[class]..name.."|r:  "..v1.icon.." "..i1,
          playerName = name,
          cooldown = v1.cooldown,
          spellName = v1.spellName
        }
      end
    end
  end
end

function LKA_RemoveFromAssignments(name)
  if(name) then
    for i,v in pairs(stuns.content) do
      if(v.playerName == name) then
        stuns.content[i] = nil;
      end
    end
    for i,v in pairs(rotation_stuns.content) do
      if(v.playerName == name) then
        rotation_stuns.content[i] = nil;
      end
    end
    for i,v in pairs(aoe_stuns.content) do
      if(v.playerName == name) then
        aoe_stuns.content[i] = nil;
      end
    end
    for i,v in pairs(slows.content) do
      if(v.playerName == name) then
        slows.content[i] = nil;
      end
    end
    for i,v in pairs(aoe_slows.content) do
      if(v.playerName == name) then
        aoe_slows.content[i] = nil;
      end
    end
    for i,v in pairs(infests.content) do
      if(v.playerName == name) then
        infests.content[i] = nil;
      end
    end
  end
end

function LKA_AnnounceDefile(person) --LKAnnounceSaved.options.defile
	if((phase == 2 and LKAnnounceSaved.options.defile.announce.enabled2) or (phase == 3 and LKAnnounceSaved.options.defile.announce.enabled3)) then
		local role = LKAnnounceSaved.options.defile.announce.role and ((talentLib:GetUnitRole(person) and string.upper(talentLib:GetUnitRole(person))) or "?")
		SendChatMessage("{skull}Defile on "..(role.." " or "")..person.."{skull}", LKAnnounceSaved.options.defile.announce["channel"..phase], nil, nil);
		if(LKAnnounceSaved.options.defile.announce.whisper) then
			SendChatMessage("{skull}DEFILE ON YOU!{skull}", "WHISPER", nil, person);
		end
	end
	if(person == UnitName("player")) then
		if(LKAnnounceSaved.options.defile.alert.self) then
			local c = LKAnnounceSaved.options.defile.alert.self_color;
			LKA_FlashScreen(c.r, c.g, c.b, c.a);
		end
	else
		if(LKAnnounceSaved.options.defile.alert.other) then
			local c = LKAnnounceSaved.options.defile.alert.other_color;
			LKA_FlashScreen(c.r, c.g, c.b, c.a);
		end
	end
end

function LKA_AnnounceInfest(num)
	if((phase == 1 and LKAnnounceSaved.options.infest.number.enabled1) or (phase == 2 and LKAnnounceSaved.options.infest.number.enabled2)) then
		local cooldown = "";
		if(LKAnnounceSaved.options.infest.cooldown["enabled"..phase]) then
			cooldown = " No cooldown available!";
			for i,v in pairs(LKAnnounceSaved.valks.infest) do
				local entry = infests.content[v];
				if(entry and UnitIsConnected(entry.playerName) and not(UnitIsDeadOrGhost(entry.playerName) or picked[entry.playerName] or cooldowns[v])) then
					LKA_debug("Cooldown "..v.." chosen for infest "..num);
					cooldown = " Cooldown: "..v;
					break;
				end
			end
		end
		SendChatMessage("Infest #"..num.." "..((num == 1 and "casting") or "coming up!")..cooldown, LKAnnounceSaved.options.infest.number["channel"..phase], nil, nil);
  end
end

function LKA_AnnounceHealingAssignment(name, occurrence)
  if(LKAnnounceSaved.valks.heal[name]) then
    SendChatMessage(name.." "..occurrence.."! Backup assignment: "..LKAnnounceSaved.valks.heal[name], LKAnnounceSaved.options.heal.disabled.channel, nil, nil);
  end
end

function LKA_HandleCombatMessage(self, event, ...)
  local timestamp, type, sourceGUID, sourceName, sourceFlags, destGUID, destName, destFlags = select(1, ...)
  if(type == "SPELL_AURA_APPLIED") then
    local announce = 0;
    local spellId, spellName, spellSchool = select(9, ...)
    -- Use the following line in game version 3.0 or higher, for previous versions use the line after
    local amount, overkill, school, resisted, blocked, absorbed, critical, glancing, crushing = select(12, ...)
    if(spellName=="Frenzy" and sourceName == "Shambling Horror") then
      if(LKAnnounceSaved.options.plague.frenzy.enabled) then
        SendChatMessage("{skull}SHAMBLING HORROR FRENZY - COOLDOWNS ON ADD TANK!!{skull}", LKAnnounceSaved.options.plague.frenzy.channel, nil, nil);
      end
    end
    if(spellName=="Enrage" and sourceName == "Shambling Horror") then
      if(LKAnnounceSaved.options.plague.enrage.enabled) then
        SendChatMessage("Shambling Horror used Enrage - Stun/Tranq/Cooldown now!", LKAnnounceSaved.options.plague.enrage.channel, nil, nil);
      end
    end
  elseif(type == "SPELL_CAST_START") then
    local spellId, spellName, spellSchool = select(9, ...)
    local amount, overkill, school, resisted, blocked, absorbed, critical, glancing, crushing = select(12, ...)
    if(spellName == "Defile") then
      pending_defile = true;
      Chronos.schedule(0.5, function()
        pending_defile = false;
      end);
    elseif(spellName == "Infest") then
      if(infest_count == 1) then
        LKA_AnnounceInfest(infest_count);
      end
      infest_count = infest_count + 1;
      Chronos.scheduleByName("Infest", 17, LKA_AnnounceInfest, infest_count);
      --[[Chronos.scheduleByName("Reset", 60, function()
        infest_count = 1;
      end);]] --Don't think this is necessary
    elseif(spellName == "Remorseless Winter") then
      infest_count = 1;
      Chronos.unscheduleByName("Infest");
			phase = phase + 1;
    end
  elseif(type == "SPELL_CAST_SUCCESS") then
    local spellId, spellName, spellSchool = select(9, ...)
    local amount, overkill, school, resisted, blocked, absorbed, critical, glancing, crushing = select(12, ...)
    if(spellName == "Defile") then
      pending_defile = false;
    elseif(spellName == "Harvest Soul") then
      local _, _, _, _, _, heroic = GetInstanceInfo();
      if(heroic == 0) then
        LKA_debug("Harvest Soul on "..destName);
        if(LKAnnounceSaved.options.heal.disabled.harvest_soul) then
          LKA_AnnounceHealingAssignment(destName, "has Harvest Soul");
        end
      else
        LKA_debug("Heroic mode; ignoring harvest souls");
      end
    else
      local key = sourceName.." - "..spellName;
      for i,v in pairs(LKAnnounceSaved.valks.stun) do
        if(v == key and stuns.content[v] and stuns.content[v].cooldown ~= 0) then
          LKA_debug(key.." used.");
          cooldowns[key] = GetTime();
          Chronos.schedule(stuns.content[v].cooldown, function()
            LKA_debug(key.." ready.");
            cooldowns[key] = nil;
          end);
        end
      end
      for i,v in pairs(LKAnnounceSaved.valks.aoestun) do
        if(v == key and aoe_stuns.content[v] and aoe_stuns.content[v].cooldown ~= 0) then
          LKA_debug(key.." used.");
          cooldowns[key] = GetTime();
          Chronos.schedule(aoe_stuns.content[v].cooldown, function()
            LKA_debug(key.." ready.");
            cooldowns[key] = nil;
          end);
        end
      end
      for i,v in pairs(LKAnnounceSaved.valks.infest) do
        if(v == key and infests.content[v] and infests.content[v].cooldown ~= 0) then
          LKA_debug(key.." used.");
          cooldowns[key] = GetTime();
          Chronos.schedule(infests.content[v].cooldown, function()
            LKA_debug(key.." ready.");
            cooldowns[key] = nil;
          end);
        end
      end
    end
  elseif(type == "SPELL_PERIODIC_DAMAGE") then
    local spellName = select(10, ...)
    if(spellName == "Necrotic Plague") then
      --A plague just damaged someone - give it time to jump, then see if it can be detected - this is only necessary 1 out of every 3 ticks, but it's easier than trying to track when a plague actually kills something
      Chronos.schedule(0.5, LKA_Calculate_Plagues);
    end
  elseif(type == "SPELL_DISPEL") then
    local extraSpellName = select(13, ...)
    if(extraSpellName == "Necrotic Plague") then
      --A plague was just dispelled off someone - give it time to jump, then see if it can be detected
      Chronos.schedule(0.5, LKA_Calculate_Plagues);
    end
  elseif(type == "SPELL_DAMAGE") then
    local spellId, spellName, spellSchool = select(9, ...)
    local amount, overkill, school, resisted, blocked, absorbed, critical, glancing, crushing = select(12, ...)
    if(spellName == "Shadow Trap" or spellName == "Ice Burst") then
      if(UnitIsPlayer(destName)) then
        trap_targets[destName] = 1;
        Chronos.scheduleByName("Traps", 0.5, LKA_Announce_Trap_Hits);
      end
    end
  elseif(type == "SPELL_SUMMON") then
    local spellId, spellName, spellSchool = select(9, ...)
    if(spellName == "Summon Val'kyr" and sourceName == "The Lich King") then
      Chronos.schedule(10, function()
        Chronos.unscheduleByName("Valks");
      end);
      if not(Chronos.isScheduledByName("Valks")) then
        wipe(picked);
        num_picked = 0;
        Chronos.scheduleRepeating("Valks", 0.5, LKA_Scan_Valks);
      end
    end
  elseif(type == "SPELL_AURA_APPLIED_DOSE") then
    --One plague merged into another plague
    --This refreshes the stack, so the old lockout timer for announcing the plague should be removed
    --LKA_Calculate_Plagues() does not need to be called because it will be triggered by the combat message that caused the merger plague to jump in the first place
    Chronos.unscheduleByName(sourceGUID);
    plague_announced[sourceGUID] = nil;
  elseif(type == "UNIT_DIED") then
    --A unit died - most of the units that die in this fight will be plagued adds in phase 1, so check plagues after every death
    Chronos.schedule(0.5, LKA_Calculate_Plagues);
    LKA_debug(destName.." ("..destGUID..") died");
    if(LKAnnounceSaved.options.heal.disabled.death) then
      LKA_AnnounceHealingAssignment(destName, "died");
    end
  end
end

function LKA_Calculate_Plagues()
  --first, construct a list of detectable horrors from the list of people who are targetting horrors - this basically reverses the targeting_horrors table by reversing the keys and values and thereby collapsing repeat values
  local horrors = {};
  for i,v in pairs(targeting_horrors) do
    --don't do any detection for horrors who have already had their current stack of plague announced
    if not(plague_announced[v]) then
      horrors[v] = i;
    end
  end
  --now we should have a table where the keys are unique GUIDs (unique horrors) and the values are corresponding UIDs
  
  --check the buffs of every relevant horror
  for GUID,targetingUID in pairs(horrors) do
    local UID = targetingUID.."target";
    local CoE = 1;
    local pTime, pCount, pTicks, pDamage;
    local frenzy = false;
    local name, count, expirationTime;
    for i=1,40 do
      name, _, _, count, _, _, expirationTime = UnitDebuff(UID,i);
      if(name == "Necrotic Plague") then
        pCount = count;
        pTime = expirationTime - GetTime();
        pTicks = math.ceil(pTime/5);
      end
      if(name == "Curse of the Elements") then
        CoE = 1.13;
      end
      if(name == "Earth and Moon") then
        CoE = 1.13;
      end
      if(name == "Ebon Plague") then
        CoE = 1.13;
      end
    end
    
    local difficulty = GetRaidDifficulty();
    local _, _, _, _, _, heroic = GetInstanceInfo();
    if(difficulty == 1) then
      pDamage = 50000;
      if(heroic == 1) then
        pDamage = 75000;
        frenzy = true;
      end
    elseif(difficulty == 2) then
      pDamage = 100000;
      if(heroic == 1) then
        pDamage = 150000;
        frenzy = true;
      end
    elseif(difficulty == 3) then
      pDamage = 75000;
      frenzy = true;
    elseif(difficulty == 4) then
      pDamage = 150000;
      frenzy = true;
    end
    
    --pCount will be null if the horror didn't have a plague
    if(pCount) then
      --Mark that this horror's plague has been announced, and set that announcement mark to wear off at the same as the plague will wear off (in pTime seconds)
      plague_announced[GUID] = true;
      LKA_debug(GUID.."'s plague announced; do not announce it again for "..pTime.." seconds");
      Chronos.scheduleByName(GUID, pTime, function()
        plague_announced[GUID] = nil;
        LKA_debug(GUID.."'s plague wore off; it is now eligible for announcements again");
      end);
      
      local horrorName = horror_names[GUID] or "Shambling Horror "..GUID;
      local diseaseDamage = pCount * pDamage * CoE;
      local health = UnitHealth(UID);
      local maxHealth = UnitHealthMax(UID);
      local frenzyHealth = UnitHealthMax(UID) * "0.20";
      local percent = (health / maxHealth) * 100;
      local ticks = math.ceil(pTime / 5);
      local remainder = pTime % 5;
      
      LKA_debug(pCount.." stacks, "..pDamage.." damage per stack("..difficulty.."), "..diseaseDamage.." total");
      
      local message = horrorName.." health: "..LKA_format_number(health).."("..format("%.1f", percent).."%)";
      local timeTillDeath = 0;
      local timeTillFrenzy = 0;
      for i = 1, ticks do
        local timeUntil = ((i - 1) * 5) + remainder;
        local projectedHealth = health - (diseaseDamage * i);
        if(projectedHealth <= 0) then
          message = message.." - Plague will kill it in "..format("%.1f", timeUntil).." seconds.";
          if(timeTillDeath ~= 0) then
            LKA_debug("Break error");
          end
          timeTillDeath = timeUntil;
          break;
        end
        if(frenzy and projectedHealth <= frenzyHealth) then
          message = message.." - It will Frenzy in "..format("%.1f", timeUntil).." seconds!";
          timeTillFrenzy = timeUntil;
        end
        if(i == ticks) then
          local projectedPercent = (projectedHealth / maxHealth) * 100;
          message = message.." - Plague will leave it at "..LKA_format_number(projectedHealth).."("..format("%.1f", projectedPercent).."%)";
        end
      end
      if(LKAnnounceSaved.options.plague.track.enabled) then
        SendChatMessage(message, LKAnnounceSaved.options.plague.track.channel);
        if(timeTillFrenzy > 0) then
          if(timeTillDeath == 0) then
            SendChatMessage(horrorName.." Frenzy in "..format("%.1f", timeTillFrenzy).."s! Plague will not kill it!", LKAnnounceSaved.options.plague.track.channel);
          elseif(timeTillDeath - timeTillFrenzy > 0) then
            SendChatMessage(horrorName.." Frenzy in "..format("%.1f", timeTillFrenzy).."s! Plague will kill it "..format("%.0f", (timeTillDeath - timeTillFrenzy)).."s later.", LKAnnounceSaved.options.plague.track.channel);
          else
            LKA_debug("Plague calculation error");
          end
        end
      end
    end
  end
end

function LKA_Announce_Trap_Hits()
  if(LKAnnounceSaved.options.trap.fly.enabled) then
    local target_num = 0;
    local targets_string = "";
    for index, value in pairs(trap_targets) do
      targets_string = targets_string..index..", ";
      target_num = target_num + 1;
    end
    targets_string = string.sub(targets_string, 0, -3);
    if(target_num == 1) then
      local gender = UnitSex(targets_string);
      if(gender == 1) then
        SendChatMessage(targets_string.."'s like a bi-ird, it'll only fly awa-ay...", "YELL", nil, nil);
      elseif(gender == 2) then
        SendChatMessage(targets_string.."'s like a bi-ird, he'll only fly awa-ay...", "YELL", nil, nil);
      elseif(gender == 3) then
        SendChatMessage(targets_string.."'s like a bi-ird, she'll only fly awa-ay...", "YELL", nil, nil);
      else
        LKA_debug("Gender error! (Oh noes!)");
      end
    else
      SendChatMessage(targets_string.." are like a bi-ird, they'll only fly awa-ay...", "YELL", nil, nil);
    end
  end
  wipe(trap_targets);
end

function LKA_Scan_Valks()
  for i=1,GetNumRaidMembers() do
    local UID = "raid"..i;
    local name = UnitName(UID);
    if(UnitInVehicle(UID) and not picked[name]) then
      LKA_debug(name.." picked up!");
      local announced = false;
      if(LKAnnounceSaved.valks.heal[name]) then
        if(LKAnnounceSaved.options.heal.disabled.valkyr) then
          LKA_AnnounceHealingAssignment(name, "picked up");
          announced = true;
        end
      end
      if not(announced) then
        if(LKAnnounceSaved.options.valkyr.picked.enabled) then
          SendChatMessage(name.." picked up!", LKAnnounceSaved.options.valkyr.picked.channel);
        end
      end
      num_picked = num_picked + 1;
      picked[UnitName(UID)] = "picked up";
    end
  end
  
  local numValks = 1;
  local diff = GetInstanceDifficulty();
  if(diff == 2 or diff == 4) then
    numValks = 3;
  end
  
  if(num_picked >= numValks) then
    Chronos.unscheduleByName("Valks");
    LKA_Handle_Valks();
  end
end

function LKA_Handle_Valks()
  LKA_debug("Handling valks");
  --Add disabled players to the picked list
  for i=1,GetNumRaidMembers() do
    local UID = "raid"..i;
    if(UnitIsDeadOrGhost(UID)) then
      picked[UnitName(UID)] = "dead";
    end
    if not(UnitIsConnected(UID)) then
      picked[UnitName(UID)] = "offline";
    end
  end
  
  --first figure out who is available for back ups
  local backupstun = {};
  for i=4,6 do
    v = stuns.content[LKAnnounceSaved.valks.stun[i]];
    if(v) then
      if not(UnitIsDeadOrGhost(v.playerName) or picked[v.playerName] or cooldowns[LKAnnounceSaved.valks.stun[i]]) then
        tinsert(backupstun, v.playerName);
      end
    end
  end
  
  local backupslow = {};
  for i=4,6 do
    v = slows.content[LKAnnounceSaved.valks.slow[i]];
    if(v) then
      if not(UnitIsDeadOrGhost(v.playerName) or picked[v.playerName] or cooldowns[LKAnnounceSaved.valks.slow[i]]) then
        tinsert(backupslow, v.playerName);
      end
    end
  end
  
  --assign people to aoe assignments
  if(LKAnnounceSaved.options.valkyr.aoestun.enabled) then
    local aoestun = nil;
    local num_aoe = 0;
    for i,v in pairs(LKAnnounceSaved.valks.aoestun) do
      local entry = aoe_stuns.content[v];
      if(entry) then
        num_aoe = num_aoe + 1;
        if not(UnitIsDeadOrGhost(entry.playerName) or picked[entry.playerName] or cooldowns[v]) then
          --elligible to stun (in raid, not dead, not picked, stun not on cooldown)
          aoestun = entry;
          break;
        end
      end
    end
    if(aoestun) then
      if(num_aoe == 1) then
        --The AoE stunner picked was the first one
        if(LKAnnounceSaved.options.valkyr.aoestun.main_enabled) then
          SendChatMessage(aoestun.playerName.." needs to AoE stun!", LKAnnounceSaved.options.valkyr.aoestun.main_channel, nil, nil);
        end
        if(LKAnnounceSaved.options.valkyr.aoestun.main_whisper) then
          SendChatMessage("AoE stun!", "WHISPER", nil, aoestun.playerName);
        end
      else
        if(LKAnnounceSaved.options.valkyr.aoestun.backup_enabled) then
          SendChatMessage(aoestun.playerName.." needs to AoE stun!", LKAnnounceSaved.options.valkyr.aoestun.backup_channel, nil, nil);
        end
        if(LKAnnounceSaved.options.valkyr.aoestun.backup_whisper) then
          SendChatMessage("AoE stun!", "WHISPER", nil, aoestun.playerName);
        end
      end
    else
      if(LKAnnounceSaved.options.valkyr.aoestun.backup_enabled) then
        SendChatMessage("No AoE stun! We're boned!", LKAnnounceSaved.options.valkyr.aoestun.backup_channel, nil, nil);
      end
    end
  end
  
  if(LKAnnounceSaved.options.valkyr.aoeslow.enabled) then
    local aoeslow = nil;
    local num_aoe = 0;
    for i,v in pairs(LKAnnounceSaved.valks.aoeslow) do
      local entry = aoe_slow.content[v];
      if(entry) then
        num_aoe = num_aoe + 1;
        if not(UnitIsDeadOrGhost(entry.playerName) or picked[entry.playerName] or cooldowns[v]) then
          --elligible to stun (in raid, not dead, not picked, stun not on cooldown)
          aoestun = entry;
          break;
        end
      end
    end
    if(aoestun) then
      if(num_aoe == 1) then
        --The AoE stunner picked was the first one
        if(LKAnnounceSaved.options.valkyr.aoeslow.main_enabled) then
          SendChatMessage(aoeslow.playerName.." needs to AoE slow!", LKAnnounceSaved.options.valkyr.aoeslow.main_channel, nil, nil);
        end
        if(LKAnnounceSaved.options.valkyr.aoeslow.main_whisper) then
          SendChatMessage("AoE slow!", "WHISPER", nil, aoestun.playerName);
        end
      else
        if(LKAnnounceSaved.options.valkyr.aoeslow.backup_enabled) then
          SendChatMessage(aoeslow.playerName.." needs to AoE slow!", LKAnnounceSaved.options.valkyr.aoeslow.backup_channel, nil, nil);
        end
        if(LKAnnounceSaved.options.valkyr.aoeslow.backup_whisper) then
          SendChatMessage("AoE slow!", "WHISPER", nil, aoestun.playerName);
        end
      end
    else
      if(LKAnnounceSaved.options.valkyr.aoeslow.backup_enabled) then
        SendChatMessage("No AoE slow! We're boned!", LKAnnounceSaved.options.valkyr.aoeslow.backup_channel, nil, nil);
      end
    end
  end
  
  --now check for picked people for each assignment and assign suitable backups
  if(LKAnnounceSaved.options.valkyr.stun.enabled) then
    for i=1,3 do
      v = stuns.content[LKAnnounceSaved.valks.stun[i]];
      if(v and (picked[v.playerName] or cooldowns[LKAnnounceSaved.valks.stun[i]])) then
        --this will only occur when someone in the picked list has the same name as someone in the assigned stuns list
        i1 = LKAnnounceSaved.valks.stun[i + 10];
        v1 = stuns.content[i1];
        local v2;
        local has_backup = false;
        if(v1) then
          v2 = v1.playerName;
          has_backup = true;
        end
        if not(v1 and not(picked[v1.playerName] or cooldowns[i1])) then
          --This could either occur if v1 doesn't exist, or if v1 is ineligible due to being picked, etc.
          if(#backupstun > 0) then
            --find the first element in the backupstun table
            for i3, v3 in pairs(backupstun) do
              has_backup = true;
              v2 = v3;
              break;
            end
          end
        end
        if(has_backup) then
          if(LKAnnounceSaved.options.valkyr.stun.backup_enabled) then
            SendChatMessage(v.playerName.." is "..(picked[v.playerName] or "on cooldown").."! "..v2.." needs to stun "..shapes.content[LKAnnounceSaved.valks.shapes[i]].chatName.."!", LKAnnounceSaved.options.valkyr.stun.backup_channel, nil, nil);
          end
          if(LKAnnounceSaved.options.valkyr.stun.backup_whisper) then
            SendChatMessage("Stun "..shapes.content[LKAnnounceSaved.valks.shapes[i]].chatName.."!", "WHISPER", nil, v2);
          end
          --this backup stun is no longer available on this round of valkyr
          tremove(backupstun, i1);
        else
          if(LKAnnounceSaved.options.valkyr.stun.backup_enabled) then
            SendChatMessage("No stun for "..shapes.content[LKAnnounceSaved.valks.shapes[i]].chatName.."! We're boned!", LKAnnounceSaved.options.valkyr.stun.backup_channel, nil, nil);
          end
        end
      elseif(v) then
        if(LKAnnounceSaved.options.valkyr.stun.main_enabled) then
          SendChatMessage(v.playerName.." needs to stun "..shapes.content[LKAnnounceSaved.valks.shapes[i]].chatName.."!", LKAnnounceSaved.options.valkyr.stun.main_channel, nil, nil);
        end
        if(LKAnnounceSaved.options.valkyr.stun.main_whisper) then
          SendChatMessage("Stun "..shapes.content[LKAnnounceSaved.valks.shapes[i]].chatName.."!", "WHISPER", nil, v.playerName);
        end
      end
    end
  end
  
  if(LKAnnounceSaved.options.valkyr.slow.enabled) then
    for i=1,3 do
      v = slows.content[LKAnnounceSaved.valks.slow[i]];
      if(v and (picked[v.playerName] or cooldowns[LKAnnounceSaved.valks.slow[i]])) then
        --this will only occur when someone in the picked list has the same name as someone in the assigned stuns list
        i1 = LKAnnounceSaved.valks.slow[i + 10];
        v1 = slow.content[i1];
        local v2;
        local has_backup = false;
        if(v1) then
          v2 = v1.playerName;
          has_backup = true;
        end
        if not(v1 and not(picked[v1.playerName] or cooldowns[i1])) then
          --This could either occur if v1 doesn't exist, or if v1 is ineligible due to being picked, etc.
          if(#backupslow > 0) then
            --find the first element in the backupslow table
            for i3, v3 in pairs(backupslow) do
              has_backup = true;
              v2 = v3;
              break;
            end
          end
        end
        if(has_backup) then
          if(LKAnnounceSaved.options.valkyr.slow.backup_enabled) then
            SendChatMessage(v.playerName.." is "..(picked[v.playerName] or "on cooldown").."! "..v2.." needs to slow "..shapes.content[LKAnnounceSaved.valks.shapes[i]].chatName.."!", LKAnnounceSaved.options.valkyr.slow.backup_channel, nil, nil);
          end
          if(LKAnnounceSaved.options.valkyr.slow.backup_whisper) then
            SendChatMessage("Slow "..shapes.content[LKAnnounceSaved.valks.shapes[i]].chatName.."!", "WHISPER", nil, v2);
          end
          --this backup stun is no longer available on this round of valkyr
          tremove(backupslow, i1);
        else
          if(LKAnnounceSaved.options.valkyr.slow.backup_enabled) then
            SendChatMessage("No slow for "..shapes.content[LKAnnounceSaved.valks.shapes[i]].chatName.."! We're boned!", LKAnnounceSaved.options.valkyr.slow.backup_channel, nil, nil);
          end
        end
      elseif(v) then
        if(LKAnnounceSaved.options.valkyr.slow.main_enabled) then
          SendChatMessage(v.playerName.." needs to slow "..shapes.content[LKAnnounceSaved.valks.shapes[i]].chatName.."!", LKAnnounceSaved.options.valkyr.slow.main_channel, nil, nil);
        end
        if(LKAnnounceSaved.options.valkyr.stun.main_whisper) then
          SendChatMessage("Slow "..shapes.content[LKAnnounceSaved.valks.shapes[i]].chatName.."!", "WHISPER", nil, v.playerName);
        end
      end
    end
  end
  
  wipe(picked);
end

function LKA_wipe_valk_assignments()
  LKAnnounceSaved.valks = {
    aoestun = {},
    aoeslow = {},
    stun = {},
    slow = {},
    infest = {},
    heal = {},
    shapes = {1, 2, 3}
  };
end

function LKA_DefaultOptions()
  LKAnnounceSaved.options = {
    trap = {
      fly = {
        enabled = true,
        channel = "YELL"
      }
    },
    plague = {
      track = {
        enabled = true,
        channel = "RAID"
      },
      enrage = {
        enabled = true,
        channel = "RAID"
      },
      frenzy = {
        enabled = true,
        channel = "RAID_WARNING"
      }
    },
    infest = {
      number = {
        enabled1 = true,
        channel1 = "RAID_WARNING",
        enabled2 = true,
        channel2 = "RAID_WARNING"
      },
      cooldown = {
        enabled1 = true,
        enabled2 = true
      }
    },
    valkyr = {
      picked = {
        enabled = false,
        channel = "RAID",
        mark = false
      },
      stun = {
        enabled = true,
        backup_enabled = true,
        backup_channel = "RAID",
        backup_whisper = true,
        main_enabled = false,
        main_channel = "RAID",
        main_whisper = false
      },
      aoestun = {
        enabled = true,
        backup_enabled = true,
        backup_channel = "RAID",
        backup_whisper = true,
        main_enabled = false,
        main_channel = "RAID",
        main_whisper = false
      },
      slow = {
        enabled = true,
        backup_enabled = true,
        backup_channel = "RAID",
        backup_whisper = true,
        main_enabled = false,
        main_channel = "RAID",
        main_whisper = false
      },
      aoeslow = {
        enabled = false,
        backup_enabled = false,
        backup_channel = "RAID",
        backup_whisper = false,
        main_enabled = false,
        main_channel = "RAID",
        main_whisper = false
      }
    },
    defile = {
      announce = {
        enabled2 = true,
        channel2 = "RAID",
        enabled3 = true,
        channel3 = "RAID",
        role = true,
        whisper = true
      },
      alert = {
        self = true,
        self_color = {r = 1, g = 0, b = 0, a = 0.3},
        other = false,
        other_color = {r = 0, g = 1, b = 0, a = 0.3}
      }
    },
    heal = {
      disabled = {
        options_list = {
          death = "Death",
          leave_raid = "Leave Raid",
          valkyr = "Valkyr (phase 2)",
          harvest_soul = "Harvest Soul (phase 3)"
        },
        channel = "RAID",
        death = true,
        leave_raid = true,
        valkyr = true,
        harvest_soul = true
      }
    }
  }
end

LKA_OnLoad();
