assert(LibStub, "LibBlizzOptions-1.0 Requires LibStub");

local MAJOR, MINOR = "LibBlizzOptions-1.0", "8 9999999" -- This hack job looks bizzare, but it will always make SVN Checkouts supercede a packaged version.
local lib = LibStub:NewLibrary(MAJOR, MINOR)

if not lib then return end

local titleFont = CreateFont("LBO_TitleFont");
titleFont:SetFont("Fonts\\FRIZQT__.TTF", 20, "OUTLINE");
titleFont:SetTextColor(0.6, 0.6, 0.9, 1);
titleFont:SetShadowColor(0.1,0.1,0.3,0.9);
titleFont:SetShadowOffset(2,-2);
titleFont:SetJustifyH("LEFT");
titleFont:SetJustifyV("BOTTOM");

local genFont = CreateFont("LBO_GeneralFont");
genFont:SetFont("Fonts\\FRIZQT__.TTF", 12);
genFont:SetShadowColor(0,0,0,0.7);
genFont:SetShadowOffset(1,-1);
genFont:SetJustifyH("LEFT");
genFont:SetJustifyV("TOP");

function lib:CreateBlizzardOptions(params)

	if type(params) ~= "table" then
		assert("Cannot Create Options Frame, Invalid Parameters provided.");
	end
	
	assert(params.Mod, "LibBlizzOptions-1.0: Parameter 'Mod' Missing. (Required)");
	assert(params.Title, "LibBlizzOptions-1.0: Parameter 'Title' Missing. (Required)");
	assert(params.Description, "LibBlizzOptions-1.0: Parameter 'Description' Missing. (Required)");
	
	local frame = lib:CreateBaseFrame(params.Mod);
	local edit = lib:CreateEditBox(frame);
	
	frame.name = params.Title;
	frame.Title:SetText(params.Title);
	frame.Description:SetText(params.Description);

	frame.Meta = lib:CreateMetaFrame(frame, params);
	
	if (params.Execute and params.ExecuteText) then
		frame.Execute = lib:CreateExecuteButton(frame, params.Execute, params.ExecuteText);
	end
	
	InterfaceOptions_AddCategory(frame);

end

function lib:CreateBaseFrame(modd)

	local frame = CreateFrame("frame", modd.."_BlizzOptions", UIParent);
	
	local title = frame:CreateFontString("$parent_Title", "ARTWORK");
	title:SetFontObject(titleFont);
	title:SetHeight(24);
	title:SetPoint("TOPLEFT", frame, "TOPLEFT", 10, -10);
	title:SetPoint("TOPRIGHT", frame, "TOPRIGHT", -10, -10);
	title:SetText("Mod Title");
	title:Show();
	frame.Title = title;
	
	local desc = frame:CreateFontString("$parent_Description", "ARTWORK");
	desc:SetFontObject(genFont);
	desc:SetTextColor(1,1,1,1);
	desc:SetHeight(75);
	desc:SetPoint("TOPLEFT", title, "BOTTOMLEFT", 10, -2);
	desc:SetPoint("TOPRIGHT", title, "BOTTOMRIGHT", 0, -2);
	desc:SetText("Mod Description. Mod Description. Mod Description. Mod Description. Mod Description. Mod Description. Mod Description. Mod Description. Mod Description. Mod Description.");
	desc:Show();
	frame.Description = desc;
	
	frame:Show();
	
	return frame;
end

function lib:CreateMetaFrame(base, params)
	local frame = CreateFrame("frame", "$parent_Meta", base);
	frame:SetPoint("TOPLEFT", base, "TOPLEFT", 45, -120);
	frame:SetPoint("TOPRIGHT", base, "TOPRIGHT", -25, -120);
	frame:SetHeight(200);
	frame:Show();
	frame.Edit = base.Edit;
	
	local metas = {}
	
	-- Version
	assert(params.Version, "LibBlizzOptions-1.0: Parameter 'Version' Missing. (Required)");
	assert(params.Author, "LibBlizzOptions-1.0: Parameter 'Author' Missing. (Required)");
	table.insert(metas, lib:CreateMetaLine(frame, "version", "Version", params.Version));
	table.insert(metas, lib:CreateMetaLine(frame, "author", "Author", params.Author));
	
	if (params.Category) then
		table.insert(metas, lib:CreateMetaLine(frame, "cats", "Category", params.Category));
	end
	
	if (params.Email) then
		table.insert(metas, lib:CreateMetaLine(frame, "email", "E-Mail", params.Email));
	end
	
	if (params.Webpage) then
		table.insert(metas, lib:CreateMetaLine(frame, "webpage", "Webpage", params.Webpage));
	end
	
	if (params.Credits) then
		table.insert(metas, lib:CreateMetaLine(frame, "credit", "Credits", params.Credits));
	end
	
	local anchor = frame;
	for i = 1,#metas do
		local m = metas[i];
		
		if (anchor == frame) then
			m:SetPoint("TOPLEFT", frame, "TOPLEFT");
			m:SetPoint("TOPRIGHT", frame, "TOPRIGHT");
		else
			m:SetPoint("TOPLEFT", anchor, "BOTTOMLEFT", 0, -1);
			m:SetPoint("TOPRIGHT", anchor, "BOTTOMRIGHT", 0, -1);
		end
		
		anchor = m;
	end
	
	return frame;
end

function lib:CreateMetaLine(base, name, title, text)
	local frame = CreateFrame("button", "$parent_"..name, base);
	frame:SetHeight(14);
	frame:SetHighlightTexture("Interface\\Buttons\\UI-Listbox-Highlight.blp");
	frame:Show();
	
	local t = frame:CreateFontString("$parent_Title", "ARTWORK");
	t:SetFontObject(genFont);
	t:SetTextColor(0.6, 0.6, 0.7, 1);
	t:SetText(title);
	t:SetPoint("TOPLEFT", frame, "TOPLEFT");
	t:SetJustifyH("RIGHT");
	t:SetHeight(14);
	t:SetWidth(75);
	t:Show();
	
	local d = frame:CreateFontString("$parent_Data", "ARTWORK");
	d:SetFontObject(genFont);
	d:SetTextColor(0.8, 0.8, 1, 1);
	d:SetText(text);
	d:SetPoint("TOPLEFT", frame, "TOPLEFT", 80, 0);
	d:SetPoint("TOPRIGHT", frame, "TOPRIGHT");
	d:SetHeight(14);
	d:Show();
	
	frame.Title = t;
	frame.Edit = base.Edit;
	frame.Data = d;
	
	frame:SetScript("OnClick", function(self)
		self.Edit:SetText(self.Data:GetText());
		lib:EditBoxFadeIn(self.Edit);
	end);
	
	return frame;
end

function lib:CreateEditBox(base)
	local eb = CreateFrame('EditBox', "$parent_DisplayText", base)
	eb:Hide()
	eb:EnableMouse(true);
	eb:SetHeight(32)
	eb:SetFontObject('GameFontHighlightSmall')

	local left = eb:CreateTexture(nil, "BACKGROUND")
	left:SetWidth(8) 
	left:SetHeight(20)
	left:SetPoint("LEFT", -5, 0)
	left:SetTexture("Interface\\Common\\Common-Input-Border")
	left:SetTexCoord(0, 0.0625, 0, 0.625)

	local right = eb:CreateTexture(nil, "BACKGROUND")
	right:SetWidth(8) 
	right:SetHeight(20)
	right:SetPoint("RIGHT", 0, 0)
	right:SetTexture("Interface\\Common\\Common-Input-Border")
	right:SetTexCoord(0.9375, 1, 0, 0.625)

	local center = eb:CreateTexture(nil, "BACKGROUND")
	center:SetHeight(20)
	center:SetPoint("RIGHT", right, "LEFT", 0, 0)
	center:SetPoint("LEFT", left, "RIGHT", 0, 0)
	center:SetTexture("Interface\\Common\\Common-Input-Border")
	center:SetTexCoord(0.0625, 0.9375, 0, 0.625)
	
	eb:SetPoint("BOTTOMLEFT", base, "BOTTOMLEFT", 20, 10);
	eb:SetPoint("BOTTOMRIGHT", base, "BOTTOMRIGHT", -20, 10);
	
	eb.FadeIn = false;
	eb.FadeOut = false;
	eb.FadeTime = 0;
	
	eb:SetScript("OnEditFocusGained", eb.HighlightText)
	eb:SetScript("OnEscapePressed", function(self) lib:EditBoxFadeOut(self); end);
	eb:SetScript("OnEnterPressed", function(self) lib:EditBoxFadeOut(self); end);
	
	base.Edit = eb;
	
	return eb;
end

function lib:EditBoxFadeIn(box)
	box.FadeTime = 0;
	
	box:Show();
	box:SetAlpha(0);
	
	box:SetScript("OnUpdate", function(self, elapsed)
		self.FadeTime = self.FadeTime + elapsed;
		
		if (self.FadeTime >= 1) then
			self:SetAlpha(1);
			self:SetScript("OnUpdate", nil);
		else
			self:SetAlpha(self.FadeTime);
		end
	end);
end

function lib:EditBoxFadeOut(box)
	box.FadeTime = 0;
	box:ClearFocus();
	box:Show();
	box:SetAlpha(1);
	
	box:SetScript("OnUpdate", function(self, elapsed)
		self.FadeTime = self.FadeTime + elapsed;
		
		if (self.FadeTime >= 1) then
			self:SetAlpha(0);
			self:Hide();
			self:SetScript("OnUpdate", nil);
		else
			self:SetAlpha(1 - self.FadeTime);
		end
	end);
end

function lib:CreateExecuteButton(base, exFunction, exText)
	local btn = CreateFrame("button", "$parent_OpenConfig", base, "OptionsButtonTemplate");
	btn:SetHeight(24);
	btn:SetPoint("LEFT", base, "LEFT", 55, 0);
	btn:SetPoint("RIGHT", base, "RIGHT", -55, 0);
	btn:SetPoint("TOP", base.Meta, "BOTTOM");
	btn:SetText(exText);
	btn:SetScript("OnClick", exFunction);
	btn:Show();
	
	return btn;
end
